#!/usr/bin/python
#
# Operations on pathnames using a Venti store.
#       - Cameron Simpson <cs@cskk.id.au> 07may2013
#

''' Assorted path related utility functions and classes.
'''

from abc import ABC, abstractmethod
import errno
import os
from os.path import dirname, join as joinpath, exists as pathexists
from stat import S_ISDIR
from cs.buffer import CornuCopyBuffer
from cs.fileutils import datafrom
from cs.logutils import error, warning
from cs.pfx import Pfx
from . import fromtext, PATHSEP

def decode_Dirent_text(text):
  ''' Accept `text`, a text transcription of a Dirent, such as from
      Dirent.textencode(), and return the corresponding Dirent.
  '''
  from .dir import _Dirent
  data = fromtext(text)
  E, offset = _Dirent.from_bytes(data)
  if offset < len(data):
    raise ValueError("%r: not all text decoded: got %r with unparsed data %r"
                     % (text, E, data[offset:]))
  return E

def dirent_dir(direntpath, do_mkdir=False):
  ''' Convert a direntpath into a Dir.
  '''
  D, name, unresolved = dirent_resolve(direntpath, do_mkdir=do_mkdir)
  if unresolved:
    raise ValueError("unresolved remaining path: %r" % (unresolved,))
  if name is not None:
    if name in D or not do_mkdir:
      D = D.chdir1(name)
    else:
      D = D.mkdir(name)
  return D

def dirent_file(direntpath, do_create=False):
  ''' Convert a direntpath into a FileDirent.
  '''
  E, name, unresolved = dirent_resolve(direntpath)
  if unresolved:
    raise ValueError("unresolved remaining path: %r" % (unresolved,))
  if name is None:
    return E
  if name in E:
    return E[name]
  if not do_create:
    raise ValueError("no such file: %s" % (direntpath,))
  raise RuntimeError("file creation not yet implemented")

def dirent_resolve(direntpath, do_mkdir=False):
  ''' Resolve `direntpath`, optionally making missing intermediate directories.
  '''
  rootD, tail = get_dirent(direntpath)
  return resolve(rootD, tail, do_mkdir=do_mkdir)

def get_dirent(direntpath):
  ''' Take `direntpath` starting with a text transcription of a Dirent and
      return the Dirent and the remaining path.
  '''
  try:
    hexpart, tail = direntpath.split('/', 1)
  except ValueError:
    hexpart = direntpath
    tail = ''
  return decode_Dirent_text(hexpart), tail

def path_split(path):
  ''' Split path into components, discarding the empty string and ".".
      The returned subparts are useful for path traversal.
  '''
  return [ subpath for subpath in path.split('/') if subpath != '' and subpath != '.' ]

def resolve(rootD, subpath, do_mkdir=False):
  ''' Descend from the Dir `rootD` via the path `subpath`.
      Return the final Dirent, its parent, and a list of unresolved path components.

      `subpath` may be a str or an array of str.
  '''
  if not rootD.isdir:
    raise ValueError("resolve: not a Dir: %s" % (rootD,))
  E = rootD
  parent = E.parent
  if isinstance(subpath, str):
    subpaths = path_split(subpath)
  else:
    subpaths = subpath
  while subpaths and E.isdir:
    name = subpaths[0]
    if name == '' or name == '.':
      # stay on this Dir
      pass
    elif name == '..':
      # go up a level if available
      if E.parent is None:
        break
      E = E.parent
    elif name in E:
      parent = E
      E = E[name]
    elif do_mkdir:
      parent = E
      E = E.mkdir(name)
    else:
      break
    subpaths.pop(0)
  return E, parent, subpaths

def walk(rootD, topdown=True, yield_status=False):
  ''' An analogue to os.walk to descend a vt Dir tree.
      Yields Dir, relpath, dirnames, filenames for each directory in the tree.
      The top directory (`rootD`) has the relpath ''.
  '''
  if not topdown:
    raise ValueError("topdown must be true, got %r" % (topdown,))
  ok = True
  # queue of (Dir, relpath)
  pending = [ (rootD, '') ]
  while pending:
    thisD, relpath = pending.pop(0)
    dirnames = thisD.dirs()
    filenames = thisD.files()
    yield thisD, relpath, dirnames, filenames
    with Pfx("walk(relpath=%r)", relpath):
      for name in reversed(dirnames):
        with Pfx("name=%r", name):
          try:
            subD = thisD.chdir1(name)
          except KeyError as e:
            if not yield_status:
              raise
            error("chdir1(%r): %s", name, e)
            ok = False
          else:
            if relpath:
              subpath = os.path.join(relpath, name)
            else:
              subpath = name
            pending.append( (subD, subpath) )
  if yield_status:
    yield ok

class DirLike(ABC):
  ''' Facilities offered by directory like objects.
  '''

  isdir = property(lambda self: True)
  isfile = property(lambda self: False)

  @staticmethod
  def check_subname(name):
    ''' Sanity check `name`, raise ValueError if invalid.
    '''
    if not name:
      raise ValueError("name may not be empty")
    if PATHSEP in name:
      raise ValueError("name may not contain PATHSEP %r: %r" % (PATHSEP, name))

  @abstractmethod
  def __contains__(self, name):
    raise NotImplementedError("no %s.__contains__" % (type(self),))

  @abstractmethod
  def __getitem__(self, name):
    raise NotImplementedError("no %s.__getitem__" % (type(self),))

  def get(self, name, default=None):
    ''' Mapping method.
    '''
    try:
      return self[name]
    except KeyError:
      return default

  @abstractmethod
  def __delitem__(self, name):
    raise NotImplementedError("no %s.__delitem__" % (type(self),))

  @abstractmethod
  def keys(self):
    ''' The names in this DirLike.
    '''
    raise NotImplementedError("no %s.__getitem__" % (type(self),))

  def __iter__(self):
    return iter(self.keys())

  def values(self):
    ''' Iterator yielding values.
    '''
    for k in self.keys():
      try:
        v = self[k]
      except KeyError:
        pass
      else:
        yield v

  def items(self):
    ''' Iterator yielding `(key, vlaue)` tuples.
    '''
    for k in self.keys():
      try:
        v = self[k]
      except KeyError:
        pass
      else:
        yield k, v

  @abstractmethod
  def lstat(self):
    ''' Return a stat object.
        Depending on the backend various fields may be None or meaningless.
        Returns None if `self` does not represent
        an existing object in the backend.
    '''
    raise NotImplementedError("no %s.stat" % (type(self),))

  @abstractmethod
  def mkdir(self, name):
    ''' Construct and return a new empty subdirectory.
    '''
    raise NotImplementedError("no %s.mkdir" % (type(self),))

  @abstractmethod
  def file_frombuffer(self, name, bfr):
    ''' Contruct a new file containing data from `bfr`.
    '''
    raise NotImplementedError("no %s.file_frombuffer" % (type(self),))

  def file_fromchunks(self, name, chunks):
    ''' Create a new file named `name` from the data in `chunks`.
    '''
    return self.file_frombuffer(name, CornuCopyBuffer(chunks))

  def resolve(self, rpath):
    ''' Resolve `rpath` relative to `self`, return resolved node or `None`.

        *WARNING*: this can walk upwards an arbitrary distance.
    '''
    node = self
    for part in rpath.split(PATHSEP):
      if not part:
        continue
      if part == '.':
        continue
      if part == '..':
        node = self.parent
        if node is None:
          break
      node = self.get(part)
      if node is None:
        break
    return node

  def walk(self):
    ''' Walk this tree.
    '''
    pending = [ (self, []) ]
    while pending:
      node, rparts = pending.pop()
      rpath = PATHSEP.join(rparts)
      dirnames = []
      filenames = []
      for name in sorted(node.keys()):
        subnode = node.get(name)
        if subnode is None:
          continue
        if subnode.isdir:
          dirnames.append(name)
        else:
          filenames.append(name)
      odirnames = set(dirnames)
      yield rpath, dirnames, filenames
      for name in dirnames:
        if name not in odirnames:
          warning(
              "walk(%s): %r: dirname %r not in original set",
              self, rpath, name)
          continue
        node = node.get(name)
        if node is None:
          continue
        pending.append( (node, rparts + [name]) )

class FileLike(ABC):
  ''' Facilities offered by file like objects.
  '''

  isdir = property(lambda self: False)
  isfile = property(lambda self: True)

  @abstractmethod
  def lstat(self):
    ''' Return a stat object.
        Depending on the backend various fields may be None or meaningless.
        Returns None if `self` does not represent
        an existing object in the backend.
    '''
    raise NotImplementedError("no %s.stat" % (type(self),))

  @abstractmethod
  def datafrom(self):
    ''' Iterator yielding natural data chunks from the file.
    '''
    raise NotImplementedError("no %s.datafrom" % (type(self),))

  def bufferfrom(self):
    ''' Return a CornuCopyBuffer presenting data from the file.
    '''
    return CornuCopyBuffer(self.datafrom())

class OSDir(DirLike):
  ''' DirLike class for an OS directory.
  '''

  def __init__(self, path):
    if not path:
      path = '.'
    DirLike.__init__(self)
    self.path = path

  def __str__(self):
    return "%s(%r)" % (type(self).__name__, self.path)

  def keys(self):
    ''' Directory entry names.
    '''
    return os.listdir(self.path)

  def __contains__(self, name):
    return name in self.keys()

  def __getitem__(self, name):
    if name not in self.keys():
      raise KeyError(name)
    namepath = joinpath(self.path, name)
    try:
      S = os.stat(namepath)
    except OSError as e:
      if e.errno == errno.ENOENT:
        raise KeyError(name)
      raise
    if S_ISDIR(S.st_mode):
      return type(self)(namepath)
    return OSFile(namepath)

  def __delitem__(self, name):
    self.check_subname(name)
    os.remove(joinpath(self.path, name))

  def lstat(self):
    with Pfx("lstat(%r)", self.path):
      try:
        return os.lstat(self.path)
      except OSError as e:
        if e.errno == errno.ENOENT:
          return None
        raise

  @property
  def parent(self):
    ''' This directory's parent.
    '''
    if PATHSEP in self.path:
      return OSDir(dirname(self.path))
    return None

  def exists(self):
    ''' Test if the directory exists.
    '''
    return pathexists(self.path)

  def create(self):
    ''' Create this directory.
    '''
    os.mkdir(self.path)

  def mkdir(self, name):
    ''' Create a subdirectory.
    '''
    if not name or PATHSEP in name:
      raise ValueError("name may not be empty or contain PATHSEP %r: %r" % (PATHSEP, name))
    subpath = joinpath(self.path, name)
    with Pfx("mkdir(%r)", subpath):
      os.mkdir(subpath)
    return OSDir(subpath)

  def file_frombuffer(self, name, bfr):
    ''' Create a new file from data from a CornuCopyBuffer.
    '''
    if not name or PATHSEP in name:
      raise ValueError("name may not be empty or contain PATHSEP %r: %r" % (PATHSEP, name))
    subpath = joinpath(self.path, name)
    with open(subpath, 'wb') as f:
      for data in bfr:
        f.write(data)
    return OSFile(subpath)

class OSFile(FileLike):
  ''' FileLike class for an OS file.
  '''

  def __init__(self, path):
    FileLike.__init__(self)
    self.path = path

  def lstat(self):
    with Pfx("lstat(%r)", self.path):
      try:
        return os.lstat(self.path)
      except OSError as e:
        if e.errno == errno.ENOENT:
          return None
        raise

  def exists(self):
    ''' Test if the file exists.
    '''
    return pathexists(self.path)

  def datafrom(self):
    ''' Yield data from the file.
    '''
    with open(self.path, 'rb') as f:
      yield from datafrom(f, 0)



#!/usr/bin/python -tt
#
# The basic flat file data store for vt blocks.
# These are kept in a directory accessed by a DataDir class.
# The file extension is .vtd
#       - Cameron Simpson <cs@cskk.id.au>
#

''' Implementation of DataFile: a file containing Block records.
'''

from enum import IntFlag
from fcntl import flock, LOCK_EX, LOCK_UN
import os
from os import SEEK_END, \
               O_CREAT, O_EXCL, O_RDONLY, O_WRONLY, O_APPEND
import sys
from zlib import compress, decompress
from cs.binary import BSUInt, BSData, PacketField
from cs.fileutils import ReadMixin, datafrom_fd
from cs.logutils import warning
from cs.resources import MultiOpenMixin
from . import Lock

DATAFILE_EXT = 'vtd'
DATAFILE_DOT_EXT = '.' + DATAFILE_EXT

class DataFlag(IntFlag):
  ''' Flag values for DataFile records.

      `COMPRESSED`: the data are compressed using zlib.compress.
  '''
  COMPRESSED = 0x01

class DataRecord(PacketField):
  ''' A data chunk file record.
  '''

  TEST_CASES = (
      (b'', b'\x01\x08x\x9c\x03\x00\x00\x00\x00\x01'),
  )

  def __init__(self, data, is_compressed=False):
    self._data = data
    self._is_compressed = is_compressed

  def __str__(self):
    return "%s(%d-bytes,%s)" % (
        type(self).__name__,
        len(self._data),
        "compressed" if self._is_compressed else "raw",
    )

  def __eq__(self, other):
    return self.data == other.data

  @classmethod
  def from_buffer(cls, bfr):
    ''' Parse a DataRecord from a buffer.
    '''
    flags = BSUInt.value_from_buffer(bfr)
    data = BSData.value_from_buffer(bfr)
    is_compressed = (flags & DataFlag.COMPRESSED) != 0
    flags &= ~DataFlag.COMPRESSED
    if flags:
      raise ValueError("unsupported flags: 0x%02x" % (flags,))
    return cls(data, is_compressed=is_compressed)

  def transcribe(self, uncompressed=False):
    ''' Transcribe this data chunk as a data record.
    '''
    data = self._data
    is_compressed = self._is_compressed
    if uncompressed:
      flags = 0x00
      if is_compressed:
        data = decompress(data)
    else:
      flags = DataFlag.COMPRESSED
      if not is_compressed:
        data = compress(data)
    yield BSUInt.transcribe_value(flags)
    yield BSData.transcribe_value(data)

  @property
  def data(self):
    ''' The uncompressed data.
    '''
    raw_data = self._data
    if self._is_compressed:
      raw_data = decompress(raw_data)
      self._data = raw_data
      self._is_compressed = False
    return raw_data

class DataFileReader(MultiOpenMixin, ReadMixin):
  ''' Read access to a data file, storing data chunks in compressed form.
      This is the usual file based persistence layer of a local Store.
  '''

  def __init__(self, pathname, lock=None):
    MultiOpenMixin.__init__(self, lock=lock)
    self.pathname = pathname
    self._rfd = None
    self._rlock = None

  def __str__(self):
    return "%s(%s)" % (type(self).__name__, self.pathname,)

  def startup(self):
    ''' Start up the DataFile: open the read and write file descriptors.
    '''
    rfd = os.open(self.pathname, O_RDONLY)
    self._rfd = rfd
    self._rlock = Lock()

  def shutdown(self):
    ''' Shut down the DataFIle: close read and write file descriptors.
    '''
    os.close(self._rfd)
    self._rfd = None
    self._rlock = None

  def datafrom(self, offset, readsize=None):
    ''' Yield data from the file starting at `offset`.
    '''
    if readsize is None:
      # Default read size.
      # This number is arbitrary, chosen on the basis that the
      # average size of blocks for random data is around 4093 bytes
      # (from vt.scan) and the size for parsed data is often much
      # smaller.
      readsize = 2048
    return datafrom_fd(self._rfd, offset, readsize)

  def fetch_record(self, offset):
    ''' Fetch a DataRecord from the supplied `offset`.
    '''
    return DataRecord.from_buffer(self.bufferfrom(offset))

  def fetch(self, offset):
    ''' Fetch the nucompressed data at `offset`.
    '''
    return self.fetch_record(offset).data

  @staticmethod
  def scanbuffer(bfr):
    ''' Generator yielding DataRecords and end offsets from a DataFile.

        Parameters:
        * `bfr`: the buffer.
        * `do_decompress`: decompress the scanned data, default False.
    '''
    while True:
      try:
        record = DataRecord.from_buffer(bfr)
      except EOFError:
        break
      yield record, bfr.offset

  def scanfrom(self, offset=0):
    ''' Generator yielding (DataRecord, post_offset) from the
        DataFile starting from `offset`, default 0.
    '''
    return self.scanbuffer(self.bufferfrom(offset))

class DataFileWriter(MultiOpenMixin):
  ''' Append access to a data file, storing data chunks in compressed form.
  '''

  def __init__(self, pathname, do_create=False, lock=None):
    MultiOpenMixin.__init__(self, lock=lock)
    self.pathname = pathname
    if do_create:
      fd = os.open(pathname, O_CREAT | O_EXCL | O_WRONLY)
      os.close(fd)
    self._wfd = None
    self._wlock = None

  def __str__(self):
    return "%s(%s)" % (type(self).__name__, self.pathname,)

  def startup(self):
    ''' Start up the DataFile: open the read and write file descriptors.
    '''
    self._wfd = os.open(self.pathname, O_WRONLY | O_APPEND)
    self._wlock = Lock()

  def shutdown(self):
    ''' Shut down the DataFIle: close read and write file descriptors.
    '''
    os.close(self._wfd)
    self._wfd = None
    self._wlock = None

  def add(self, data):
    ''' Append a chunk of data to the file, return the store start
        and end offsets.

        The fcntl.flock function is used to hold an OS level lock
        for the duration of the write to support shared use of the
        file.
    '''
    bs = bytes(DataRecord(data))
    wfd = self._wfd
    with self._wlock:
      try:
        flock(wfd, LOCK_EX)
      except OSError:
        is_locked = False
      else:
        is_locked = True
      offset = os.lseek(wfd, 0, SEEK_END)
      written = os.write(wfd, bs)
      # notice short writes, which should never happen with a regular file...
      while written < len(bs):
        warning(
            "%s: tried to write %d bytes but only wrote %d, retrying",
            self, len(bs), written)
        bs = bs[written:]
        written = os.write(wfd, bs)
      if is_locked:
        flock(wfd, LOCK_UN)
    return offset, offset + len(bs)

if __name__ == '__main__':
  from .datafile_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# INdex classes.
# - Cameron Simpson <cs@cskk.id.au>
#

''' An index is a mapping of hashcodes => bytes-records.
    This module supports several backends and a mechanism for choosing one.
'''

from contextlib import contextmanager
from os.path import exists as pathexists
from cs.logutils import warning, info
from cs.pfx import Pfx
from cs.resources import MultiOpenMixin
from . import Lock
from .hash import HashCodeUtilsMixin

_CLASSES = []
_BY_NAME = {}

def class_names():
  ''' Return the index class names.
  '''
  return _BY_NAME.keys()

def class_by_name(indexname):
  ''' Return an index class from its name.
  '''
  return _BY_NAME[indexname]

def choose(basepath, preferred_indexclass=None):
  ''' Choose an indexclass from a `basepath` with optional preferred indexclass.
  '''
  global _CLASSES
  global _BY_NAME
  if preferred_indexclass is not None:
    if isinstance(preferred_indexclass, str):
      indexname = preferred_indexclass
      try:
        preferred_indexclass = _BY_NAME[indexname]
      except KeyError:
        warning("ignoring unknown indexclass name %r", indexname)
        preferred_indexclass = None
  indexclasses = list(_CLASSES)
  if preferred_indexclass is not None and preferred_indexclass.is_supported():
    indexclasses.insert( (preferred_indexclass.NAME, preferred_indexclass) )
  for indexname, indexclass in indexclasses:
    if not indexclass.is_supported():
      continue
    indexpath = indexclass.pathof(basepath)
    if pathexists(indexpath):
      return indexclass
  for indexname, indexclass in indexclasses:
    if not indexclass.is_supported():
      continue
    return indexclass
  raise ValueError(
      "no supported index classes available: tried %r"
      % (indexclasses,))

class _Index(HashCodeUtilsMixin, MultiOpenMixin):

  def __init__(self, basepath, hashclass, decode, lock=None):
    ''' Initialise an _Index instance.

        Parameters:
        * `basepath`: the base path to the index; the index itself
          is at `basepath`.SUFFIX
        * `decode`: function to decode a binary index record into the
          return type instance
        * `lock`: optional mutex, passed to MultiOpenMixin.__init__.
    '''
    MultiOpenMixin.__init__(self, lock=lock)
    self.basepath = basepath
    self.hashclass = hashclass
    self.decode = decode

  @classmethod
  def pathof(cls, basepath):
    ''' Construct the path to the index file.
    '''
    return '.'.join((basepath, cls.SUFFIX))

  @property
  def path(self):
    ''' The path to the index file.
    '''
    return self.pathof(self.basepath)

class LMDBIndex(_Index):
  ''' LMDB index for a DataDir.
  '''

  NAME = 'lmdb'
  SUFFIX = 'lmdb'
  MAP_SIZE = 1024 * 1024 * 1024

  def __init__(self, lmdbpathbase, hashclass, decode, lock=None):
    _Index.__init__(self, lmdbpathbase, hashclass, decode, lock=lock)
    self._lmdb = None
    self._resize_needed = False
    # Locking around transaction control logic.
    self._txn_lock = Lock()
    # Lock preventing activity which cannot occur while a transaction is
    # current. This is primarily for database reopens, as when the
    # LMDB map_size is raised.
    self._txn_idle = Lock()     # available if no transactions are in progress
    self._txn_blocked = Lock()  # available if new transactions may commence
    self._txn_count = 0

  @classmethod
  def is_supported(cls):
    ''' Test whether this index class is supported by the Python environment.
    '''
    try:
      import lmdb
    except ImportError:
      return False
    return True

  def startup(self):
    ''' Start up the index.
    '''
    self.map_size = 10240   # self.MAP_SIZE
    self._open_lmdb()

  def shutdown(self):
    ''' Shut down the index.
    '''
    with self._txn_idle:
      self.flush()
      self._lmdb.close()
      self._lmdb = None

  def _open_lmdb(self):
    import lmdb
    self._lmdb = lmdb.Environment(
        self.path,
        subdir=True, readonly=False,
        metasync=False, sync=False,
        writemap=True, map_async=True,
        map_size=self.map_size,
    )

  def _embiggen_lmdb(self, new_map_size=None):
    if new_map_size is None:
      new_map_size = self.map_size * 2
    self.map_size = new_map_size
    info("change LMDB map_size to %d", self.map_size)
    # reopen the database
    self._lmdb.sync()
    self._lmdb.close()
    self._open_lmdb()

  @contextmanager
  def _txn(self, write=False):
    ''' Context manager wrapper for an LMDB transaction which tracks active transactions.
    '''
    with self._txn_blocked:
      with self._txn_lock:
        resize_needed = self._resize_needed
        if resize_needed:
          self._resize_needed = False
      if resize_needed:
        # wait for existing transactions to finish
        with self._txn_idle:
          self._embiggen_lmdb()
      with self._txn_lock:
        count = self._txn_count
        count += 1
        self._txn_count = count
        if count == 1:
          # mark transactions as in progress
          self._txn_idle.acquire()
    try:
      yield self._lmdb.begin(write=write)
    finally:
      # logic mutex
      with self._txn_lock:
        count = self._txn_count
        count -= 1
        self._txn_count = count
        if count == 0:
          # mark all transactions as complete
          self._txn_idle.release()

  def flush(self):
    ''' Flush outstanding data to the index.
    '''
    # no force=True param?
    self._lmdb.sync()

  def __iter__(self):
    mkhash = self.hashclass.from_hashbytes
    with self._txn() as txn:
      cursor = txn.cursor()
      for hashcode in cursor.iternext(keys=True, values=False):
        yield mkhash(hashcode)

  def items(self):
    ''' Yield `(hashcode,record)` from index.
    '''
    mkhash = self.hashclass.from_hashbytes
    with self._txn() as txn:
      cursor = txn.cursor()
      for hashcode, record in cursor.iternext(keys=True, values=True):
        yield mkhash(hashcode), self.decode(record)

  def _get(self, hashcode):
    with self._txn() as txn:
      return txn.get(hashcode)

  def __contains__(self, hashcode):
    return self._get(hashcode) is not None

  def __getitem__(self, hashcode):
    record = self._get(hashcode)
    if record is None:
      raise KeyError(hashcode)
    return self.decode(record)

  def get(self, hashcode, default=None):
    ''' Get and decode the record for `hashcode`.
        Return None for missing `hashcode`.
    '''
    entry = self._get(hashcode)
    if entry is None:
      return default
    return self.decode(entry)

  def __setitem__(self, hashcode, entry):
    import lmdb
    record = entry.encode()
    while True:
      try:
        with self._txn(write=True) as txn:
          txn.put(hashcode, record, overwrite=True)
          txn.commit()
      except lmdb.MapFullError as e:
        info("%s", e)
        self._resize_needed = True
      else:
        return

class GDBMIndex(_Index):
  ''' GDBM index for a DataDir.
  '''

  NAME = 'gdbm'
  SUFFIX = 'gdbm'

  def __init__(self, lmdbpathbase, hashclass, decode, lock=None):
    _Index.__init__(self, lmdbpathbase, hashclass, decode, lock=lock)
    self._gdbm = None

  @classmethod
  def is_supported(cls):
    ''' Test whether this index class is supported by the Python environment.
    '''
    try:
      import dbm.gnu
    except ImportError:
      return False
    return True

  def startup(self):
    ''' Start the index: open dbm, allocate lock.
    '''
    import dbm.gnu
    with Pfx(self.path):
      self._gdbm = dbm.gnu.open(self.path, 'cf')
    self._gdbm_lock = Lock()
    self._written = False

  def shutdown(self):
    ''' Shutdown the index.
    '''
    self.flush()
    with self._gdbm_lock:
      self._gdbm.close()
      self._gdbm = None
      del self._gdbm_lock

  def flush(self):
    ''' Flush the index: sync the gdbm.
    '''
    if self._written:
      with self._gdbm_lock:
        if self._written:
          self._gdbm.sync()
          self._written = False

  def __iter__(self):
    mkhash = self.hashclass.from_hashbytes
    with self._gdbm_lock:
      hashcode = self._gdbm.firstkey()
    while hashcode is not None:
      yield mkhash(hashcode)
      self.flush()
      with self._gdbm_lock:
        hashcode = self._gdbm.nextkey(hashcode)

  def __contains__(self, hashcode):
    with self._gdbm_lock:
      return hashcode in self._gdbm

  def __getitem__(self, hashcode):
    with self._gdbm_lock:
      entry = self._gdbm[hashcode]
    return self.decode(entry)

  def get(self, hashcode, default=None):
    ''' Get and decode the record for `hashcode`.
        Return None for missing `hashcode`.
    '''
    with self._gdbm_lock:
      entry = self._gdbm.get(hashcode, None)
    if entry is None:
      return default
    return self.decode(entry)

  def __setitem__(self, hashcode, value):
    entry = value.encode()
    with self._gdbm_lock:
      self._gdbm[hashcode] = entry
      self._written = True

class KyotoIndex(_Index):
  ''' Kyoto Cabinet index.
      Notably this uses a B+ tree for the index and thus one can
      traverse from one key forwards and backwards, which supports
      the coming Store synchronisation processes.
  '''

  NAME = 'kyoto'
  SUFFIX = 'kct'

  def __init__(self, lmdbpathbase, hashclass, decode, lock=None):
    _Index.__init__(self, lmdbpathbase, hashclass, decode, lock=lock)
    self._kyoto = None

  @classmethod
  def is_supported(cls):
    ''' Test whether this index class is supported by the Python environment.
    '''
    try:
      import kyotocabinet
    except ImportError:
      return False
    return True

  def startup(self):
    ''' Open the index.
    '''
    from kyotocabinet import DB
    self._kyoto = DB()
    self._kyoto.open(self.path, DB.OWRITER | DB.OCREATE)

  def shutdown(self):
    ''' Close the index.
    '''
    self._kyoto.close()
    self._kyoto = None

  def flush(self):
    ''' Flush pending updates to the index.
    '''
    try:
      self._kyoto.synchronize(hard=False)
    except TypeError:
      self._kyoto.synchronize()

  def __len__(self):
    return self._kyoto.count()

  def __contains__(self, hashcode):
    return self._kyoto.check(hashcode) >= 0

  def get(self, hashcode):
    ''' Get and decode the record for `hashcode`.
        Return None for missing `hashcode`.
    '''
    record = self._kyoto.get(hashcode)
    if record is None:
      return None
    return self.decode(record)

  def __getitem__(self, hashcode):
    entry = self.get(hashcode)
    if entry is None:
      raise IndexError(str(hashcode))
    return entry

  def __setitem__(self, hashcode, value):
    self._kyoto[hashcode] = value.encode()

  def hashcodes_from(self, start_hashcode=None, reverse=False):
    ''' Generator yielding the keys from the index
        in order starting with optional `start_hashcode`.

        Parameters:
        * `start_hashcode`: the starting hashcode; if missing or None,
          iteration starts with the first key in the index
        * `reverse`: iterate backward if true, otherwise forward
    '''
    hashclass = self.hashclass
    cursor = self._kyoto.cursor()
    if reverse:
      if cursor.jump_back(start_hashcode):
        yield hashclass.from_hashbytes(cursor.get_key())
        while cursor.step_back():
          yield hashclass.from_hashbytes(cursor.get_key())
    else:
      if cursor.jump(start_hashcode):
        yield hashclass.from_hashbytes(cursor.get_key())
        while cursor.step():
          yield hashclass.from_hashbytes(cursor.get_key())
    cursor.disable()

def register(indexclass, indexname=None, priority=False):
  ''' Register a new `indexclass`, making it known.

      Parameters:
      * `indexclass`: the index class
      * `indexname`: the index class name, default from indexclass.NAME
      * `priority`: if true, prepend the class to the _CLASSES list otherwise append
  '''
  if indexname is None:
    indexname = indexclass.NAME
  if indexname in _BY_NAME:
    raise ValueError(
        'cannot register index class %s: indexname %r already registered to %s'
        % (indexclass, indexname, _BY_NAME[indexname]))
  _BY_NAME[indexname] = indexclass
  if priority:
    _CLASSES.insert(0, (indexclass.NAME, indexclass))
  else:
    _CLASSES.append((indexclass.NAME, indexclass))

for klass in LMDBIndex, KyotoIndex, GDBMIndex:
  if klass.is_supported():
    register(klass)



#!/usr/bin/python
#
# Datafile tests.
# - Cameron Simpson <cs@cskk.id.au>
#

import os
import sys
import random
import tempfile
import unittest
try:
  import kyotocabinet
except ImportError:
  kyotocabinet = None
from cs.binary_tests import _TestPacketFields
##from cs.debug import thread_dump
from .randutils import rand0, randblock
from . import datafile
from .datafile import DataFileReader, DataFileWriter
# from .hash_tests import _TestHashCodeUtils
# TODO: run _TestHashCodeUtils on DataDirs as separate test suite?

# arbitrary limit
MAX_BLOCK_SIZE = 16383
RUN_SIZE = 100

class TestDataFilePacketFields(_TestPacketFields, unittest.TestCase):

  def setUp(self):
    self.module = datafile

class TestDataFile(unittest.TestCase):

  def setUp(self):
    random.seed()
    tfd, pathname = tempfile.mkstemp(prefix="datafile-test", suffix=".vtd", dir='.')
    os.close(tfd)
    self.pathname = pathname
    self.wdatafile = DataFileWriter(pathname)
    self.wdatafile.open()
    self.rdatafile = DataFileReader(pathname)
    self.rdatafile.open()

  def tearDown(self):
    self.wdatafile.close()
    self.rdatafile.close()
    os.remove(self.pathname)

  # TODO: tests:
  #   scan datafile

  def test00store1(self):
    ''' Save a single block.
    '''
    self.wdatafile.add(randblock(rand0(MAX_BLOCK_SIZE + 1)))

  def test01fetch1(self):
    ''' Save and the retrieve a single block.
    '''
    data = randblock(rand0(MAX_BLOCK_SIZE + 1))
    self.wdatafile.add(data)
    data2 = self.rdatafile.fetch(0)
    self.assertEqual(data, data2)

  def test02randomblocks(self):
    ''' Save RUN_SIZE random blocks, close, retrieve in random order.
    '''
    blocks = {}
    for n in range(RUN_SIZE):
      with self.subTest(put_block_n=n):
        data = randblock(rand0(MAX_BLOCK_SIZE + 1))
        offset, offset2 = self.wdatafile.add(data)
        blocks[offset] = data
    offsets = list(blocks.keys())
    random.shuffle(offsets)
    for n, offset in enumerate(offsets):
      with self.subTest(shuffled_offsets_n=n, offset=offset):
        data = self.rdatafile.fetch(offset)
        self.assertTrue(data == blocks[offset])

def multitest_suite(testcase_class, *a, **kw):
  suite = unittest.TestSuite()
  for method_name in dir(testcase_class):
    if method_name.startswith('test'):
      ta = list(a) + [method_name]
      suite.addTest(testcase_class(*ta, **kw))
  return suite

def selftest(argv):
  unittest.main(__name__, None, argv)
  ##suite = unittest.TestSuite()
  ##suite.addTest(multitest_suite(TestDataFile))
  ##suite.addTest(TestDataFilePacketFields())
  ##runner = unittest.TextTestRunner(failfast=True, verbosity=2)
  ##runner.run(suite)
  ##if False:
  ##  import cProfile
  ##  cProfile.runctx('unittest.main(__name__, None, argv)', globals(), locals())
  ##else:
  ##  unittest.main(__name__, None, argv)
  ##thread_dump()

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python

''' Functions and classes relating to Blocks, which are data chunk references.

    All Blocks derive from the base class _Block.

    The following Block types are provided:

    HashCodeBlock: a reference to data by its hashcode.

    LiteralBlock: a Block containing its literal data, used when
      the serialisation of a HashCodeBlock exceeds the length of the
      data.

    RLEBlock: a run length encoded Block for repeated byte values,
      particularly long runs of the NUL byte.

    SubBlock: a Block whose data is a subspan of another Block,
      particularly used when new data are blockified from part of an
      existing run of Blocks where the part does not start or end on
      a Block boundary.

    IndirectBlock: a Block whose data is the concatenation of a
      sequence of subsidiary Bloacks, which themselves may also be
      IndirectBlocks. This is how larger files are composed of finite
      sized Blocks. An IndirectBlock is internally constructed as
      a wrapper for another Block whose data are the serialisation
      of the subblock references.

    All Blocks have a .span attribute, which is the length of the
    data they encompass. For all "leaf" Blocks this value is same
    same as the length of their "direct" data, but for IndirectBlocks
    this is the sum of the .span values of their subblocks.
'''

from __future__ import print_function
from abc import ABC
from enum import IntEnum, unique as uniqueEnum
from functools import lru_cache
import sys
from icontract import require
from cs.binary import (
    PacketField, BSUInt, BSData,
    flatten as flatten_transcription
)
from cs.buffer import CornuCopyBuffer
from cs.lex import texthexify, untexthexify, get_decimal_value
from cs.logutils import warning
from cs.pfx import Pfx
from cs.py.func import prop
from cs.threads import locked
from cs.x import X
from . import defaults, totext, RLock
from .hash import HashCode
from .transcribe import Transcriber, register as register_transcriber, parse

F_BLOCK_INDIRECT = 0x01     # indirect block
F_BLOCK_TYPED = 0x02        # block type provided, otherwise BT_HASHCODE
F_BLOCK_TYPE_FLAGS = 0x04   # type-specific flags follow type

@uniqueEnum
class BlockType(IntEnum):
  ''' Block type codes used in binary serialisation.
  '''
  BT_INDIRECT = -1          # never gets transcribed
  BT_HASHCODE = 0           # default type: hashref
  BT_RLE = 1                # run length encoding: span octet
  BT_LITERAL = 2            # span raw-data
  BT_SUBBLOCK = 3           # a SubBlock of another Block

class BlockRecord(PacketField):
  ''' PacketField support binary parsing and transcription of blockrefs.
  '''

  TEST_CASES = (
      # zero length hashcode block
      # note that the hashcode doesn't match that of b''
      b'\x17\0\0' + b'\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0',
  )

  @staticmethod
  def value_from_buffer(bfr):
    ''' Decode a Block reference from a buffer.

        Format is:

            BS(length)
            BS(flags)
              0x01 indirect blockref
              0x02 typed: type follows, otherwise BT_HASHCODE
              0x04 type flags: per type flags follow type
            BS(span)
            [BS(type)]
            [BS(type_flags)]
            union {
              type BT_HASHCODE: hash
              type BT_RLE: octet-value (repeat span times to get data)
              type BT_LITERAL: raw-data (span bytes)
              type BT_SUBBLOCK: suboffset, super block
            }

        Even though this is all decodable without the leading length
        we use a leading length so that future encodings do not
        prevent parsing any following data.
    '''
    raw_encoding = BSData.value_from_buffer(bfr)
    blockref_bfr = CornuCopyBuffer.from_bytes(raw_encoding)
    flags = BSUInt.value_from_buffer(blockref_bfr)
    is_indirect = bool(flags & F_BLOCK_INDIRECT)
    is_typed = bool(flags & F_BLOCK_TYPED)
    has_type_flags = bool(flags & F_BLOCK_TYPE_FLAGS)
    unknown_flags = flags & ~(F_BLOCK_INDIRECT|F_BLOCK_TYPED|F_BLOCK_TYPE_FLAGS)
    if unknown_flags:
      raise ValueError(
          "unexpected flags value (0x%02x) with unsupported flags=0x%02x"
          % (flags, unknown_flags))
    span = BSUInt.value_from_buffer(blockref_bfr)
    if is_indirect:
      # With indirect blocks, the span is of the implied data, not
      # the referenced block's data. Therefore we build the referenced
      # block with a span of None and store the span in the indirect
      # block.
      ispan = span
      span = None
    # block type, default BT_HASHCODE
    if is_typed:
      block_type = BlockType(BSUInt.value_from_buffer(blockref_bfr))
    else:
      block_type = BlockType.BT_HASHCODE
    if has_type_flags:
      type_flags = BSUInt.value_from_buffer(blockref_bfr)
      if type_flags:
        warning("nonzero type_flags: 0x%02x", type_flags)
    else:
      type_flags = 0x00
    # instantiate type specific block ref
    if block_type == BlockType.BT_HASHCODE:
      hashcode = HashCode.from_buffer(blockref_bfr)
      B = HashCodeBlock(hashcode=hashcode, span=span)
    elif block_type == BlockType.BT_RLE:
      octet = blockref_bfr.take(1)
      B = RLEBlock(span, octet)
    elif block_type == BlockType.BT_LITERAL:
      data = blockref_bfr.take(span)
      B = LiteralBlock(data)
    elif block_type == BlockType.BT_SUBBLOCK:
      suboffset = BSUInt.value_from_buffer(blockref_bfr)
      superB = BlockRecord.value_from_buffer(blockref_bfr)
      # wrap inner Block in subspan
      B = SubBlock(superB, suboffset, span)
    else:
      raise ValueError("unsupported Block type 0x%02x" % (block_type,))
    if is_indirect:
      B = _IndirectBlock(B, span=ispan)
    if not blockref_bfr.at_eof():
      warning(
          "unparsed data (%d bytes) follow Block %s",
          len(raw_encoding) - blockref_bfr.offset, B)
    return B

  @staticmethod
  def transcribe_value(B):
    ''' Transcribe this Block, the inverse of value_from_buffer.
    '''
    transcription = []
    is_indirect = B.indirect
    span = B.span
    if is_indirect:
      # aside from the span, everything else comes from the superblock
      ##X("INDIRECT: B(%s) => B.superblock(%s)", B, B.superblock)
      B = B.superblock
    block_type = B.type
    assert block_type >= 0, "block_type(%s) => %d" % (B, B.type)
    block_typed = block_type != BlockType.BT_HASHCODE
    flags = (
        ( F_BLOCK_INDIRECT if is_indirect else 0 )
        | ( F_BLOCK_TYPED if block_typed else 0 )
        | 0     # no F_BLOCK_TYPE_FLAGS
    )
    transcription.append(BSUInt.transcribe_value(flags))
    transcription.append(BSUInt.transcribe_value(span))
    if block_typed:
      transcription.append(BSUInt.transcribe_value(block_type))
    # no block_type_flags
    if block_type == BlockType.BT_HASHCODE:
      transcription.append(B.hashcode.transcribe_b())
    elif block_type == BlockType.BT_RLE:
      transcription.append(B.octet)
    elif block_type == BlockType.BT_LITERAL:
      transcription.append(B.data)
    elif block_type == BlockType.BT_SUBBLOCK:
      ##X("TRANSCRIBE: B=%r", B)
      transcription.append(BSUInt.transcribe_value(B.offset))
      transcription.append(BlockRecord.transcribe_value(B.superblock))
    else:
      raise ValueError("unsupported Block type 0x%02x: %s" % (block_type, B))
    return BSData(b''.join(flatten_transcription(transcription))).transcribe()

def isBlock(o):
  ''' Test if an object `o` is a subinstance of `_Block`.
  '''
  return isinstance(o, _Block)

class _Block(Transcriber, ABC):

  def __init__(self, block_type, span):
    self.type = block_type
    if span is not None:
      if not isinstance(span, int) or span < 0:
        raise ValueError("invalid span: %r" % (span,))
      self.span = span
    self.indirect = False
    self.blockmap = None
    self._lock = RLock()

  def __eq__(self, oblock):
    if self is oblock:
      return True
    if self.span != oblock.span:
      # different lengths, no match
      return False
    # see if we can compare hashes
    h1 = getattr(self, 'hashcode', None)
    h2 = getattr(oblock, 'hashcode', None)
    if h1 is not None and h2 is not None and h1 == h2:
      if (
          self.indirect and oblock.indirect
          or not self.indirect and not oblock.indirect
      ):
        # same hashes and indirectness: same data content
        return True
    # see if both blocks are direct blocks
    if not self.indirect and not oblock.indirect:
      # directly compare data otherwise
      # TODO: this may be expensive for some Block types?
      #       some kind of rolling buffer compare required
      return self.get_spanned_data() == oblock.get_spanned_data()
    # one of the blocks is indirect: walk both blocks comparing leaves
    # we could do this by stuffing one into a buffer but we still
    # want to do direct leaf comparisons when the leaves are aligned,
    # as that may skip a data fetch
    leaves1 = self.leaves
    leaves2 = oblock.leaves
    offset = 0      # amount already compared
    offset1 = 0     # offset of start of leaf1
    offset2 = 0     # offset of start of leaf2
    leaf2 = None
    for leaf1 in leaves1:
      # skip empty leaves
      leaf1len = len(leaf1)
      if leaf1len == 0:
        continue
      end1 = offset1 + leaf1len
      while offset < end1:
        # still more bytes in leaf1 needing comparison
        # fetch leaf2 if required
        while leaf2 is None:
          try:
            leaf2 = next(leaves2)
          except StopIteration:
            # oblock is short
            return False
          if len(leaf2) == 0:
            leaf2 = None
        # compare leaves if aligned
        # this will sidestep any data fetch if they both have hashcodes
        if offset == offset1 and offset == offset2 and len(leaf1) == len(leaf2):
          if leaf1 != leaf2:
            return False
          # identical, advance to end of block
          offset = end1
          leaf2 = None
          continue
        cmplen = min(offset1, offset2) - offset
        if cmplen < 0:
          raise RuntimeError("cmplen(%d) < 0: offset=%d, offset1=%d, offset2=%d"
                             % (cmplen, offset, offset1, offset2))
        # we can defer fetching the data until now
        data1 = leaf1.data
        data2 = leaf2.data
        if ( data1[offset - offset1:offset - offset1 + cmplen]
             != data2[offset - offset2:offset - offset2 + cmplen]
           ):
          return False
        end2 = offset2 + len(data2)
        offset += cmplen
        if offset > end1 or offset > end2:
          raise RuntimeError(
              "offset advanced beyond end of leaf1 or leaf2:"
              " offset=%d, end(leaf1)=%d, end(leaf2)= %d"
              % ( offset, end1, end2))
        if offset >= end2:
          # leaf2 consumed, discard
          leaf2 = None
    # check that there are no more nonempty leaves in leaves2
    while leaf2 is None:
      try:
        leaf2 = next(leaves2)
      except StopIteration:
        break
      else:
        if len(leaf2) == 0:
          leaf2 = None
    # data are identical if we have consumed all of leaves1 an all of leaves2
    return leaf2 is None

  def encode(self):
    ''' Binary transcription of this Block via BlockRecord.
    '''
    return b''.join(BlockRecord.transcribe_value_flat(self))

  def __getitem__(self, index):
    ''' Return specified direct data.
    '''
    return self.get_spanned_data()[index]

  def __len__(self):
    ''' len(Block) is the length of the encompassed data.
    '''
    return self.span

  def get_spanned_data(self):
    ''' Collect up all the data of this Block and return a single bytes instance.

        This is painfully named because it may be very expensive.
    '''
    chunks = list(self.datafrom())
    if not chunks:
      return b''
    if len(chunks) == 1:
      return chunks[0]
    return b''.join(chunks)

  def matches_data(self, odata):
    ''' Check supplied bytes `odata` against this Block's hashcode.
        NB: _not_ defined on indirect Blocks to avoid mistakes.
    '''
    try:
      h = self.hashcode
    except AttributeError:
      return self.data == odata
    return h == h.from_chunk(odata)

  @prop
  def leaves(self):
    ''' Return the leaf (direct) blocks.
    '''
    if self.indirect:
      for B in self.subblocks:
        yield from B.leaves
    elif self.span > 0:
      yield self

  @locked
  def get_blockmap(self, force=False, blockmapdir=None):
    ''' Get the blockmap for this block, creating it if necessary.
        `force`: if true, create a new blockmap anyway; default: False
        `blockmapdir`: directory to hold persistent block maps
    '''
    if force:
      blockmap = None
    else:
      blockmap = self.blockmap
    if blockmap is None:
      warning("making blockmap for %s", self)
      from .blockmap import BlockMap
      if blockmapdir is None:
        blockmapdir = defaults.S.blockmapdir
      self.blockmap = blockmap = BlockMap(self, blockmapdir=blockmapdir)
    return blockmap

  def bufferfrom(self, offset=0, **kw):
    ''' Return a CornuCopyBuffer presenting data from the Block.
    '''
    return CornuCopyBuffer(self.datafrom(start=offset, **kw), offset=offset)

  @require(lambda self, start, end: 0 <= start <= end <= len(self))
  def slices(self, start, end, no_blockmap=False):
    ''' Return an iterator yielding (Block, start, len) tuples
        representing the leaf data covering the supplied span `start`:`end`.

        The iterator may end early if the span exceeds the Block data.
    '''
    if self.indirect:
      if not no_blockmap:
        # use the blockmap to access the data if present
        blockmap = self.blockmap
        if blockmap:
          yield from blockmap.slices(start, end - start)
          return
      offset = 0
      for B in self.subblocks:
        sublen = len(B)
        if start <= offset + sublen:
          substart = max(0, start - offset)
          subend = min(sublen, end - offset)
          yield from B.slices(substart, subend)
        offset += sublen
        if offset >= end:
          break
    else:
      # a leaf Block
      if start < len(self):
        yield self, start, min(end, len(self))

  @require(lambda self, start, end: 0 <= start <= end <= len(self))
  def top_slices(self, start, end):
    ''' Return an iterator yielding (Block, start, len) tuples
        representing the uppermost Blocks spanning `start:end`.

        The originating use case is to support providing minimal
        Block references required to assemble a new indirect Block
        consisting of data from this Block comingled with updated
        data without naively layering deeper levels of Block
        indirection with every update phase.
    '''
    if self.indirect:
      offset = 0        # the absolute index of the left edge of subblock B
      for B in self.subblocks:
        sublen = len(B)
        substart = max(0, start - offset)
        subend = min(sublen, end - offset)
        if substart < subend:
          if subend - substart == sublen:
            yield B, 0, sublen
          else:
            for subslice in B.top_slices(substart, subend):
              yield subslice
        # advance the offset to account for this subblock
        offset += sublen
        if offset >= end:
          break
    else:
      # a leaf Block
      if start < len(self):
        yield self, start, min(end, len(self))

  @require(lambda self, start, end: 0 <= start <= end <= len(self))
  def top_blocks(self, start, end):
    ''' Yield existing high level blocks and new partial Blocks
        covering a portion of this Block,
        for constructing a new minimal top block.
    '''
    for B, Bstart, Bend in self.top_slices(start, end):
      if Bstart == 0 and Bend == len(B):
        # an extant high level block
        yield B
      else:
        # should be a new partial block
        if B.indirect:
          raise RuntimeError(
              "got slice for partial Block %s start=%r end=%r"
              " but Block is indirect! should be a partial leaf"
              % (B, Bstart, Bend))
        yield SubBlock(B, Bstart, Bend - Bstart)

  @require(lambda self, start, end: 0 <= start <= end <= len(self))
  def spliced(self, start, end, new_block):
    ''' Generator yielding Blocks producing the data
        from `self` with the range `start:end`
        replaced by the data from `new_block`.
    '''
    if start == len(self):
      yield self
    else:
      yield from self.top_blocks(0, start)
    yield new_block
    if end < len(self):
      yield from self.top_blocks(end, len(self))

  @require(lambda self, start, end: 0 <= start <= end <= len(self))
  def splice(self, start, end, new_block):
    ''' Return a new Block consisting of `self` with the span
        `start:end` replaced by the data from `new_block`.
    '''
    from .blockify import top_block_for
    return top_block_for(self.spliced(start, end, new_block))

  def textencode(self):
    ''' Transcribe this Block's binary encoding as text.
        TODO: Obsolete, remove.
    '''
    return totext(self.encode())

  def open(self, mode="rb"):
    ''' Open the block as a file.
    '''
    if mode == 'rb':
      from .file import ROBlockFile
      return ROBlockFile(self)
    if mode == 'w+b':
      from .file import RWBlockFile
      return RWBlockFile(backing_block=self)
    raise ValueError(
        "unsupported open mode, expected 'rb' or 'w+b', got: %r" % (mode,))

  def pushto(self, S2, Q=None, runstate=None):
    ''' Push this Block and any implied subblocks to the Store `S2`.

        Parameters:
        * `S2`: the secondary Store to receive Blocks
        * `Q`: optional preexisting Queue, which itself should have
          come from a .pushto targetting the Store `S2`.
        * `runstate`: optional RunState used to cancel operation

        If `Q` is supplied, this method will return as soon as all
        the relevant Blocks have been pushed i.e. possibly before
        delivery is complete. If `Q` is not supplied, a new Queue
        is allocated; after all Blocks have been pushed the Queue
        is closed and its worker waited for.

        TODO: optional `no_wait` parameter to control waiting,
        default False, which would support closing the Queue but
        not waiting for the worker completion. This is on the premise
        that the final Store shutdown of `S2` will wait for outstanding
        operations anyway.
    '''
    S1 = defaults.S
    if Q is None:
      # create a Queue and a worker Thread
      Q, T = S1.pushto(S2)
    else:
      # use an existing Queue, no Thread to wait for
      T = None
    Q.put(self)
    if self.indirect:
      # recurse, reusing the Queue
      for subB in self.subblocks:
        if runstate and runstate.cancelled:
          warning("pushto(%s) cancelled", self)
          break
        subB.pushto(S2, Q, runstate=runstate)
    if T:
      Q.close()
      T.join()

@lru_cache(maxsize=1024*1024, typed=True)
def get_HashCodeBlock(hashcode):
  ''' Caching constructor for HashCodeBlocks of known code.
  '''
  if hashcode is None:
    raise ValueError("invlaid hashcode, may not be None")
  return HashCodeBlock(hashcode=hashcode)

class HashCodeBlock(_Block):
  ''' A Block reference based on a Store hashcode.
  '''

  transcribe_prefix = 'B'

  def __init__(self, hashcode=None, data=None, added=False, span=None, **kw):
    ''' Initialise a BT_HASHCODE Block or IndirectBlock.

        A HashCodeBlock always stores its hashcode directly.
        If `data` is supplied, store it and compute or check the hashcode.
        If `span` is not None, store it. Otherwise compute it on
          demand from the data, fetching that if necessary.

        NB: The data are not kept in memory; fetched on demand.
        `added`: if true, do not add the data to the current Store.
    '''
    if data is None:
      if hashcode is None:
        raise ValueError("one of data or hashcode must be not-None")
    else:
      # when constructing an indirect block, span != len(data)
      if span is None:
        span = len(data)
      if added:
        # Block already Stored, just require presupplied hashcode
        if hashcode is None:
          raise ValueError("added=%s but no hashcode supplied" % (added,))
      else:
        h = defaults.S.add(data)
        if hashcode is None:
          hashcode = h
        elif h != hashcode:
          raise ValueError(
              "supplied hashcode %r != saved hash for data (%r : %r)"
              % (hashcode, h, data))
    self._data = data
    self._span = None
    _Block.__init__(self, BlockType.BT_HASHCODE, span=span, **kw)
    self.hashcode = hashcode

  def get_direct_data(self):
    ''' Return the direct data of this Block, fetching it if necessary.
    '''
    with self._lock:
      data = self._data
      if data is None:
        data = self._data = defaults.S[self.hashcode]
    return data

  @prop
  def span(self):
    ''' Return the data length, computing it from the data if required.
    '''
    _span = self._span
    if _span is None:
      self._span = _span = len(self._data)
    return _span

  @span.setter
  def span(self, newspan):
    ''' Set the span of the data encompassed by this HashCodeBlock.
    '''
    if newspan < 0:
      raise ValueError("%s: set .span: invalid newspan=%s" % (self, newspan))
    if self._span is None:
      self._span = newspan
    else:
      warning("setting .span a second time")
      if newspan != self._span:
        raise RuntimeError(
            "%s: tried to change .span from %s to %s"
            % (self, self._span, newspan))
      else:
        raise RuntimeError("SECOND UNEXPECTED")

  def datafrom(self, start=None, end=None):
    ''' Generator yielding data from `start:end`.
    '''
    if start is not None and start < 0:
      raise ValueError("invalid start=%s" % (start,))
    if end is not None and end < 0:
      raise ValueError("invalid end=%s" % (end,))
    bs = self.get_direct_data()
    if start is None:
      if end is None:
        yield bs
        return
      raise ValueError("start is None but end=%s" % (end,))
    if end is not None and end < start:
      raise ValueError("end(%s) < start(%s)" % (end, start))
    if start == 0:
      if end is None:
        yield bs
      else:
        yield bs[:end]
    else:
      yield bs[start:end]

  def transcribe_inner(self, T, fp):
    m = {'hash': self.hashcode}
    if self._span is not None:
      m['span'] = self._span
    return T.transcribe_mapping(m, fp)

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    m, offset = T.parse_mapping(s, offset, stopchar)
    span = m.pop('span', None)
    hashcode = m.pop('hash')
    if m:
      raise ValueError("unexpected fields: %r" % (m,))
    B = cls(hashcode=hashcode, span=span)
    return B, offset

register_transcriber(HashCodeBlock, ('B', 'IB'))

def Block(*, hashcode=None, data=None, span=None, added=False):
  ''' Factory function for a Block.
  '''
  if data is None:
    if span is None:
      raise ValueError('data and span may not both be None')
    B = get_HashCodeBlock(hashcode)
  else:
    if span is None:
      span = len(data)
    elif span != len(data):
      raise ValueError("span(%d) does not match data (%d bytes)"
                       % (span, len(data)))
    if len(data) > 32:
      B = HashCodeBlock(data=data, hashcode=hashcode, span=span, added=added)
    else:
      B = LiteralBlock(data=data)
  return B

def IndirectBlock(subblocks=None, hashcode=None, span=None, force=False):
  ''' Factory function for an indirect Block.

      Indirect blocks may be initialised in two ways:

      The first way is specified by supplying the `subblocks`
      parameter, an iterable of Blocks to be referenced by this
      IndirectBlock. The referenced Blocks are encoded and assembled
      into the data for this Block.

      The second way is to supplying the `hashcode` and `span` for
      an existing Stored block, whose content is used to initialise
      an IndirectBlock is with a hashcode and a span indicating the
      length of the data encompassed by the block speified by the
      hashcode; the data of that Block can be decoded to obtain the
      reference Blocks for this IndirectBlock.

      As an optimisation, unless `force` is true: if `subblocks`
      is empty a direct Block for b'' is returned; if `subblocks`
      has just one element then that element is returned.

      TODO: allow data= initialisation, to decode raw iblock data.

  '''
  ## TODO: A direct or single byte Block should be an RLEBlock;
  ##   this breaks our implementation of .hashcode - need to see if we
  ##   can not require it - check use cases.
  if subblocks is None:
    # hashcode specified
    if hashcode is None:
      raise ValueError("one of subblocks or hashcode must be supplied")
    if span is None:
      raise ValueError("no span supplied with hashcode %s" % (hashcode,))
    B = get_HashCodeBlock(hashcode)
  else:
    # subblocks specified
    if hashcode is not None:
      raise ValueError("only one of hashocde and subblocks may be supplied")
    if isinstance(subblocks, _Block):
      subblocks = (subblocks,)
    elif isinstance(subblocks, bytes):
      subblocks = (Block(data=subblocks),)
    else:
      subblocks = tuple(subblocks)
    spans = [ subB.span for subB in subblocks ]
    subspan = sum(spans)
    if span is None:
      span = subspan
    elif span != subspan:
      raise ValueError("span(%d) does not match subblocks (totalling %d)"
                       % (span, subspan))
    if not force:
      if not subblocks:
        return Block(data=b'')
      if len(subblocks) == 1:
        return subblocks[0]
    superBdata = b''.join(subB.encode() for subB in subblocks)
    B = HashCodeBlock(data=superBdata)
  return _IndirectBlock(B, span=span)

class _IndirectBlock(_Block):

  transcribe_prefix = 'I'

  def __init__(self, superB, span=None):
    if superB.indirect:
      raise ValueError("superB may not be indirect: superB=%s" % (superB,))
    super().__init__(BlockType.BT_INDIRECT, 0)
    self.indirect = True
    self.superblock = superB
    if span is None:
      span = sum(subB.span for subB in self.subblocks)
    self.span = span
    self.hashcode = superB.hashcode
    self._data = None
    self._subblocks = None

  @prop
  @locked
  def subblocks(self):
    ''' The immediate subblocks of this indirect block.
    '''
    blocks = self._subblocks
    if blocks is None:
      blocks = self._subblocks = tuple(
          BlockRecord.parse_buffer_values(self.superblock.bufferfrom()))
    return blocks

  def transcribe_inner(self, T, fp):
    ''' Transcribe "span:Block".
    '''
    fp.write(str(self.span))
    fp.write(':')
    T.transcribe(self.superblock, fp=fp)

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    ''' Parse "span:Block"
    '''
    span, offset2 = get_decimal_value(s, offset)
    if s[offset2] != ':':
      raise ValueError("offset %d: missing colon after span(%d)" % (offset2, span))
    offset = offset2 + 1
    superB, offset = parse(s, offset, T)
    return cls(superB, span), offset

  def datafrom(self, start=0, end=None):
    ''' Yield data from a point in the Block.
    '''
    if end is None:
      end = self.span
    if start >= end:
      return
    block_cache = defaults.S.block_cache
    if block_cache:
      try:
        bm = block_cache[self.hashcode]
      except KeyError:
        pass
      else:
        filled = bm.filled
        if filled > start:
          maxlength = end - start
          for bs in bm.datafrom(start, maxlength=maxlength):
            yield bs
            start += len(bs)
            assert start <= end
    if start < end:
      for B, Bstart, Bend in self.slices(start, end):
        assert not B.indirect
        yield B.get_direct_data()[Bstart:Bend]

class RLEBlock(_Block):
  ''' An RLEBlock is a Run Length Encoded block of `span` bytes
      all of a specific value, typically NUL.
  '''

  transcribe_prefix = 'RLE'

  def __init__(self, span, octet, **kw):
    if isinstance(octet, int):
      octet = bytes((octet,))
    elif not isinstance(octet, bytes):
      raise TypeError(
          "octet should be an int or a bytes instance but is %s: %r"
          % (type(octet), octet))
    if len(octet) != 1:
      raise ValueError("len(octet):%d != 1" % (len(octet),))
    _Block.__init__(self, BlockType.BT_RLE, span=span, **kw)
    self.octet = octet

  def get_direct_data(self):
    ''' The full RLEBlock.
    '''
    return self.octet * self.span

  def datafrom(self, start=None, end=None):
    ''' Yield the data from `start` to `end`.
    '''
    if start is None:
      start = 0
    if end is None:
      end = self.span
    if end > self.span:
      end = self.span
    length = end - start
    if length < 0:
      raise ValueError("end(%s) < start(%s)" % (end, start))
    yield self.octet * length

  def transcribe_inner(self, T, fp):
    return T.transcribe_mapping({'span': self.span, 'octet': self.octet}, fp)

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    m = T.parse_mapping(s, offset, stopchar)
    span = m.pop('span')
    octet = m.pop('octet')
    if m:
      raise ValueError("unexpected fields: %r" % (m,))
    return cls(span, octet), offset

register_transcriber(RLEBlock)

class LiteralBlock(_Block):
  ''' A LiteralBlock is for data too short to bother hashing and Storing.
  '''

  transcribe_prefix = 'LB'

  def __init__(self, data, **kw):
    _Block.__init__(self, BlockType.BT_LITERAL, span=len(data), **kw)
    self.data = data

  def transcribe_inner(self, T, fp):
    ''' Transcribe the block data in texthexified form.
    '''
    fp.write(texthexify(self.data))

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    ''' Parse the interior of the transcription: texthexified data.
    '''
    endpos = s.find(stopchar, offset)
    if endpos < offset:
      raise ValueError("stopchar %r not found" % (stopchar,))
    data = untexthexify(s[offset:endpos])
    return cls(data), endpos

  def get_direct_data(self):
    ''' Return the direct data of this Block>
    '''
    return self.data

  def datafrom(self, start=0, end=None):
    if start < 0 or end is not None and start > end:
      raise ValueError("invalid start=%s (end=%s)" % (start, end))
    if end is None:
      end = self.span
    yield self.data[start:end]

register_transcriber(LiteralBlock)

def SubBlock(superB, suboffset, span, **kw):
  ''' Factory for SubBlocks.
      Returns origin Block if suboffset==0 and span==len(superB).
      Returns am empty LiteralBlock if the span==0.
  '''
  with Pfx("SubBlock(suboffset=%d,span=%d,superB=%s)", suboffset, span, superB):
    # check offset and span here because we trust them later
    if suboffset < 0 or suboffset > len(superB):
      raise ValueError("suboffset out of range")
    if span < 0 or suboffset + span > len(superB):
      raise ValueError("span(%d) out of range" % (span,))
    if span == 0:
      ##warning("span==0, returning empty LiteralBlock")
      return LiteralBlock(b'')
    if suboffset == 0 and span == len(superB):
      ##warning("covers full Block, returning original")
      return superB
    if isinstance(superB, _SubBlock):
      return _SubBlock(superB.superblock, suboffset + superB.offset, span)
    return _SubBlock(superB, suboffset, span, **kw)

class _SubBlock(_Block):
  ''' A SubBlock is a view into another block.
      A SubBlock may not be empty and may not cover the whole of its superblock.
  '''

  transcribe_prefix = 'SubB'

  def __init__(self, superB, suboffset, span, **kw):
    with Pfx("_SubBlock(suboffset=%d, span=%d)[len(superB)=%d]",
             suboffset, span, len(superB)):
      if suboffset < 0 or suboffset >= len(superB):
        raise ValueError('suboffset out of range 0-%d: %d' % (len(superB)-1, suboffset))
      if span < 0 or suboffset+span > len(superB):
        raise ValueError(
            'span must be nonnegative and less than %d (suboffset=%d, len(superblock)=%d): %d'
            % (len(superB)-suboffset, suboffset, len(superB), span))
      if suboffset == 0 and span == len(superB):
        raise RuntimeError('tried to make a SubBlock spanning all of of SuperB')
      _Block.__init__(self, BlockType.BT_SUBBLOCK, span, **kw)
      self.superblock = superB
      self.offset = suboffset

  def get_direct_data(self):
    ''' The direct data are the spanned data.
    '''
    return self.get_spanned_data()

  def datafrom(self, start=0, end=None):
    ''' Yield the data from this Block between `start` and `end`.
    '''
    if start < 0 or (end is not None and end < 0):
      raise ValueError("invalid start(%s) or end(%s)" % (start, end))
    start = self.offset + start
    if end is None:
      end = self.span
    end = self.offset + end
    return self.superblock.datafrom(start, end)

  def __getitem__(self, index):
    if isinstance(index, slice):
      return self.data[index]
    if index < 0 or index >= self.span:
      raise IndexError("index %d outside span %d" % (index, self.span))
    return self.superblock[self.offset+index]

  def transcribe_inner(self, T, fp):
    return T.transcribe_mapping({
        'block': self.superblock,
        'offset': self.offset,
        'span': self.span
    }, fp)

  @classmethod
  def parse_inner(cls, T, s, offset, stopchar, prefix):
    offset, block, suboffset, subspan = T.parse_mapping(
        s, offset, stopchar,
        required=('block', 'offset', 'span'))
    return cls(block, suboffset, subspan), offset

register_transcriber(_SubBlock)

def verify_block(B, recurse=False, S=None):
  ''' Perform integrity checks on the Block `B`, yield error messages.
  '''
  if S is None:
    S = defaults.S
  try:
    hashcode = B.hashcode
  except AttributeError:
    hashcode = None
  else:
    if hashcode not in S:
      yield str(B), "hashcode not in %s" % (S,)
    else:
      # compare the direct data versus the hashcode
      if B.indirect:
        bfr = B.superblock.datafrom()
      else:
        bfr = B.datafrom()
      hashdata = b''.join(bfr)
      # hash the data using the matching hash function
      data_hashcode = hashcode.hashfunc(hashdata)
      if hashcode != data_hashcode:
        yield str(B), "hashcode(%s) does not match hashfunc of data(%s)" \
                 % (hashcode, data_hashcode)
      Sdata = S[hashcode]
      if Sdata != hashdata:
        yield str(B), "Block hashdata != S[%s]" % (hashcode,)
  if B.indirect:
    if recurse:
      for subB in B.subblocks:
        yield from verify_block(subB, recurse=True, S=S)
  Blength = sum(len(chunk) for chunk in B.datafrom())
  if B.span != Blength:
    X("VERIFY BLOCK %s: B.span=%d, len(data)=%d",
      B, B.span, Blength)
    yield str(B), "span(%d) != len(data:%d)" % (B.span, Blength)

if __name__ == '__main__':
  from .block_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# File tests.
#       - Cameron Simpson <cs@cskk.id.au>
#

import sys
import unittest
from cs.fileutils import BackedFile_TestMethods
from . import defaults
from .blockify import blockify, top_block_for
from .store import MappingStore
from .file import RWBlockFile

class Test_RWFile(unittest.TestCase, BackedFile_TestMethods):

  def setUp(self):
    self.store_dict = {}
    self.S = MappingStore("Test_RWFile", self.store_dict)
    defaults.pushStore(self.S)
    # construct test backing block
    with open(__file__, "rb") as fp:
      self.backing_text = fp.read()
    self.vt_block = top_block_for(blockify([self.backing_text]))
    self.vt_file = RWBlockFile(self.vt_block)
    self.backed_fp = self.vt_file._file

  def tearDown(self):
    self.vt_file.close()
    defaults.popStore()

  def test_flush(self):
    self.test_BackedFile()
    B2 = self.vt_file.sync()
    self.backed_fp = self.vt_file._file
    self.assertEqual(B2, self.vt_file._backing_block)
    bfp = self.backed_fp
    self.assertEqual(bfp.front_range.end, 0, "bfp(id=%d).front_range.end should be 0, range is: %s" % (id(bfp), bfp.front_range,))

def selftest(argv):
  unittest.main(__name__, None, argv, failfast=True)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# Cache tests.
# - Cameron Simpson <cs@cskk.id.au> 25aug2015
#

import sys
import unittest
from .store import MappingStore
from .store_tests import TestStore
from .cache import FileCacheStore

class TestFileCacheStore(TestStore, unittest.TestCase):
  ''' Test the unit tests for FileCacheStore.
  '''

  def _init_Store(self):
    self.fastS = MappingStore("TestCacheStore.fastS", {}, hashclass=self.hashclass)
    self.slowS = MappingStore("TestCacheStore.slowS", {}, hashclass=self.hashclass)
    self.S = FileCacheStore("TestCacheStore.S", self.fastS, self.slowS, hashclass=self.hashclass)

def selftest(argv):
  ''' Run the unit tests.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# Filesystem checks.
#   - Cameron Simpson <cs@cskk.id.au> 11feb2017
#

from cs.logutils import error, warning, info
from cs.pfx import Pfx
from cs.x import X
from . import defaults
from .block import HashCodeBlock, LiteralBlock
from .paths import walk

def fsck_dir(rootD):
  ''' Inspect a Dir for correctness.
  '''
  ok = True
  for step in walk(rootD, yield_status=True):
    if isinstance(step, bool):
      if not step:
        error("errors from walk")
        ok = False
      break
    D, relpath, dirnames, filenames = step
    with Pfx(relpath):
      for filename in sorted(filenames):
        with Pfx("filename=%r", filename):
          try:
            E = D[filename]
          except KeyError as e:
            error("not in D")
          else:
            B = E.block
            if not fsck_Block(B):
              error("E.block fails fsck")
              ok = False
      for dirname in sorted(dirnames):
        with Pfx("dirname=%r", dirname):
          try:
            subD = D[dirname]
          except KeyError as e:
            error("not in D")
          else:
            B = subD.block
            if not fsck_Block(B):
              error("E.block fails fsck")
              ok = False
  return ok

def fsck_Block(B):
  ''' Check Block.
  '''
  ok = False
  with defaults.S as S:
    if isinstance(B, HashCodeBlock):
      hashcode = B.hashcode
      with Pfx("%s", hashcode):
        try:
          data = S[hashcode]
        except KeyError as e:
          error("not in Store %s", S)
        else:
          ok = True
          if B.indirect:
            suboffset = 0
            for i, subB in enumerate(B.subblocks):
              with Pfx("subblocks[%d]:%d..%d", i, suboffset, suboffset+len(subB)):
                if not fsck_Block(subB):
                  ok = False
                suboffset += len(subB)
            if suboffset != len(B):
              error("len(B)=%d, sum(len(subblocks))=%d", len(B), suboffset)
          else:
            if len(B) != len(data):
              error("len(B)=%d, len(data)=%d", len(B), len(data))
              ok = False
    elif isinstance(B, LiteralBlock):
      ok = True
      data = B.data
      if len(B) != len(data):
        error("len(B)=%d, len(data)=%d", len(B), len(data))
        ok = False
    else:
      error("unsupported Block type: %s", B.__class__)
  return ok



#!/usr/bin/python -tt
#
# TCP and UNIX socket client/server code.
# - Cameron Simpson <cs@cskk.id.au> 07dec2007
#

''' Support for connections over TCP and UNIX domain sockets.
'''

import os
from socket import socket, AF_INET, AF_UNIX
from socketserver import TCPServer, UnixStreamServer, \
    ThreadingMixIn, StreamRequestHandler
import sys
from threading import Thread
from cs.excutils import logexc
from cs.logutils import info, warning
from cs.pfx import Pfx
from cs.py.func import prop
from cs.queues import MultiOpenMixin
from cs.resources import RunStateMixin
from cs.socketutils import OpenSocket
from cs.x import X
from . import defaults
from .stream import StreamStore

class _SocketStoreServer(MultiOpenMixin, RunStateMixin):
  ''' The basis for TCPStoreServer and UNIXSocketStoreServer.
  '''

  def __init__(self, *, exports=None, runstate=None, local_store=None):
    ''' Initialise the server.

        Parameters:
        * `exports`: optional mapping of str=>Store
        * `runstate`: option control RunState
        * `local_store`: optional default Store

        `exports` is a mapping of Stores which this server may present;
        the default Store has key `''`.

        If `local_store` is not None, `exports['']` is set to
        `local_store` otherwise to `defaults.S`.
        It is an error to provide both `local_store` and a prefilled
        `exports['']`.
    '''
    if exports is None:
      exports = {}
    if local_store is not None:
      if '' in exports:
        raise ValueError(
            "both local_store=%s and exports['']=%s provided"
            % (local_store, exports['']))
      exports[''] = local_store
    if '' not in exports:
      exports[''] = defaults.S
    MultiOpenMixin.__init__(self)
    RunStateMixin.__init__(self, runstate=runstate)
    self.exports = exports
    self.S = exports['']
    self.socket_server = None
    self.socket_server_thread = None
    self.runstate.notify_start.add(lambda rs: self.open())
    self.runstate.notify_end.add(lambda rs: self.close())
    self.runstate.notify_cancel.add(lambda rs: self.shutdown_now())

  def __str__(self):
    return "%s[%s](S=%s)" % (type(self), self.runstate.state, self.S)

  def startup(self):
    ''' Start up the server.
    '''
    self.socket_server_thread = Thread(
        name="%s(%s)[server-thread]" % (type(self), self.S),
        target=self.socket_server.serve_forever,
        kwargs={'poll_interval': 0.5})
    self.socket_server_thread.daemon = False
    self.socket_server_thread.start()

  def shutdown(self):
    ''' Shut down the server.
    '''
    if self.socket_server:
      self.socket_server.shutdown()
    self.socket_server_thread.join()
    self.socket_server = None

  def shutdown_now(self):
    ''' Issue closes until all current opens have been consumed.
    '''
    if self.socket_server:
      self.socket_server.shutdown()
      self.socket_server = None

  def flush(self):
    ''' Flush the backing Store.
    '''
    self.S.flush()

  def join(self):
    ''' Wait for the server thread to exit.
    '''
    self.socket_server_thread.join()

class _ClientConnectionHandler(StreamRequestHandler):
  ''' Handler for a connection to the server.
  '''

  def setup(self):
    super().setup()
    self.store_server = self.server.store_server

  @prop
  def exports(self):
    ''' Return the exports mapping.
    '''
    return self.store_server.exports

  @logexc
  def handle(self):
    # the local Store starts as the current default Store
    self.S = self.store_server.S
    remoteS = StreamStore(
        "server-StreamStore(local=%s)" % self.S,
        OpenSocket(self.request, False),
        OpenSocket(self.request, True),
        local_store=self.S,
        exports=self.exports,
    )
    remoteS.startup()
    remoteS.join()
    remoteS.shutdown()

class _TCPServer(ThreadingMixIn, TCPServer):

  def __init__(self, store_server, bind_addr):
    with Pfx(
        "%s.__init__(store_server=%s, bind_addr=%r)",
        type(self).__name__, store_server, bind_addr
    ):
      TCPServer.__init__(self, bind_addr, _ClientConnectionHandler)
      self.bind_addr = bind_addr
      self.store_server = store_server

  def __str__(self):
    return "%s(%s,%s)" % (type(self), self.bind_addr, self.store_server,)

class TCPStoreServer(_SocketStoreServer):
  ''' A threading TCPServer that accepts connections from TCPClientStores.
  '''

  def __init__(self, bind_addr, **kw):
    super().__init__(**kw)
    self.bind_addr = bind_addr
    self.socket_server = _TCPServer(self, bind_addr)

class TCPClientStore(StreamStore):
  ''' A Store attached to a remote Store at `bind_addr`.
  '''

  def __init__(self, name, bind_addr, addif=False, **kw):
    if name is None:
      name = "%s(bind_addr=%r)" % (self.__class__.__name__, bind_addr)
    self.sock_bind_addr = bind_addr
    self.sock = None
    StreamStore.__init__(
        self, name, None, None,
        addif=addif, connect=self._tcp_connect,
        **kw
    )

  def shutdown(self):
    StreamStore.shutdown(self)
    if self.sock:
      self.sock.close()
      self.sock = None

  def _tcp_connect(self):
    with Pfx("connect %r", self.sock_bind_addr):
      X("TCP CONNECT to %r ...", self.sock_bind_addr)
      if self.sock:
        warning("self.sock already set, leaking: %s", self.sock)
      # TODO: IPv6 support
      self.sock = socket(AF_INET)
      try:
        self.sock.connect(self.sock_bind_addr)
      except:
        self.sock.close()
        self.sock = None
        raise
      X("TCP CONNECT to %r: CONNECTED", self.sock_bind_addr)
      return OpenSocket(self.sock, False), OpenSocket(self.sock, True)

  @logexc
  def _packet_disconnect(self, conn):
    ''' On disconnect, close the socket as well.
    '''
    super()._packet_disconnect(conn)
    sock = self.sock
    if sock:
      self.sock = None
      sock.close()

class _UNIXSocketServer(ThreadingMixIn, UnixStreamServer):

  def __init__(self, store_server, socket_path, exports=None):
    with Pfx(
        "%s.__init__(store_server=%s, socket_path=%r)",
        type(self), store_server, socket_path
    ):
      UnixStreamServer.__init__(self, socket_path, _ClientConnectionHandler)
      self.store_server = store_server
      self.socket_path = socket_path
      self.exports = exports

  def __str__(self):
    return "%s(store_server=%s,socket_path=%s,exports=%r)" \
        % (type(self), self.store_server, self.socket_path, self.exports)

class UNIXSocketStoreServer(_SocketStoreServer):
  ''' A threading UnixStreamServer that accepts connections from UNIXSocketClientStores.
  '''

  def __init__(self, socket_path, **kw):
    super().__init__(**kw)
    self.socket_path = socket_path
    self.socket_server = _UNIXSocketServer(self, socket_path)

  def shutdown(self):
    super().shutdown()
    os.remove(self.socket_path)

class UNIXSocketClientStore(StreamStore):
  ''' A Store attached to a remote Store at `socket_path`.
  '''

  def __init__(self, name, socket_path, addif=False, **kw):
    if name is None:
      name = "%s(socket_path=%r)" % (self.__class__.__name__, socket_path)
    self.socket_path = socket_path
    self.sock = None
    StreamStore.__init__(
        self, name, None, None,
        addif=addif, connect=self._unixsock_connect,
        **kw
    )

  def shutdown(self):
    StreamStore.shutdown(self)
    if self.sock:
      self.sock.close()

  def _unixsock_connect(self):
    with Pfx("connect %r", self.socket_path):
      info("UNIX SOCKET CONNECT to %r", self.socket_path)
      assert not self.sock, "self.sock=%s" % (self.sock,)
      self.sock = socket(AF_UNIX)
      with Pfx("connect(%r)", self.socket_path):
        try:
          self.sock.connect(self.socket_path)
        except OSError:
          self.sock.close()
          self.sock = None
          raise
      return OpenSocket(self.sock, False), OpenSocket(self.sock, True)

if __name__ == '__main__':
  from .socket_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# Stream protocol for stores.
#       - Cameron Simpson <cs@cskk.id.au> 06dec2007
#
# TODO: SYNC, to wait for pending requests before returning
#

''' Protocol for accessing Stores over a stream connection.
'''

from __future__ import with_statement
from enum import IntEnum
import sys
import time
from cs.binary import PacketField, EmptyField, Packet, BSString, BSUInt
from cs.buffer import CornuCopyBuffer
from cs.excutils import logexc
from cs.logutils import warning, error
from cs.packetstream import PacketConnection
from cs.pfx import Pfx
from cs.resources import ClosedError
from cs.result import CancellationError
from cs.threads import locked
from .hash import (
    decode as hash_decode,
    decode_buffer as hash_from_buffer,
    HASHCLASS_BY_NAME,
    HashCodeField,
)
from .pushpull import missing_hashcodes_by_checksum
from .store import StoreError, BasicStoreSync

class RqType(IntEnum):
  ''' Packet opcode values.
  '''
  ADD = 0               # data -> hashcode
  GET = 1               # hashcode -> data
  CONTAINS = 2          # hashcode->Boolean
  FLUSH = 3             # flush local and remote servers
  HASHCODES = 4         # (hashcode,length) -> hashcodes
  HASHCODES_HASH = 5    # (hashcode,length) -> hashcode of hashcodes

class StreamStore(BasicStoreSync):
  ''' A Store connected to a remote Store via a PacketConnection.
      Optionally accept a local store to facilitate bidirectional activities
      or simply to implement the server side.
  '''

  def __init__(
      self, name, recv, send,
      *,
      addif=False,
      connect=None,
      local_store=None,
      exports=None,
      **kw
  ):
    ''' Initialise the Stream Store.

        Parameters:
        * `name`: the Store name.
        * `recv`: inbound binary stream.
          If this is an `int` it is taken to be an OS file descriptor,
          otherwise it should be a `cs.buffer.CornuCopyBuffer`
          or a file like object with a `read1` or `read` method.
        * `send`: outbound binary stream.
          If this is an `int` it is taken to be an OS file descriptor,
          otherwise it should be a file like object with `.write(bytes)`
          and `.flush()` methods.
          For a file descriptor sending is done via an os.dup() of
          the supplied descriptor, so the caller remains responsible
          for closing the original descriptor.
        * `addif`: optional mode causing .add to probe the peer for
          the data chunk's hash and to only submit a ADD request
          if the block is missing; this is a bandwith optimisation
          at the expense of latency.
        * `connect`: if not None, a function to return `recv` and `send`.
          If specified, the `recv` and `send` parameters must be None.
        * `local_store`: optional local Store for serving requests from the peer.
        * `exports`: a mapping of name=>Store providing requestable Stores

        Other keyword arguments are passed to BasicStoreSync.__init__.
    '''
    super().__init__('StreamStore:%s' % (name,), **kw)
    self.mode_addif = addif
    self._local_store = local_store
    self.exports = exports
    if local_store is not None:
      if local_store.hashclass is not self.hashclass:
        raise ValueError("local_store.hashclass %s is not self.hashclass %s"
                         % (local_store.hashclass, self.hashclass))
    # parameters controlling connection hysteresis
    self._conn_attempt_last = 0.0
    self._conn_attempt_delay = 1.0
    if connect is None:
      # set up protocol on existing stream
      # no reconnect facility
      self._conn = conn = self._packet_connection(recv, send)
      # arrange to disassociate if the channel goes away
      conn.notify_recv_eof.add(self._packet_disconnect)
      conn.notify_send_eof.add(self._packet_disconnect)
    else:
      # defer protocol setup until needed
      if recv is not None or send is not None:
        raise ValueError("connect is not None and one of recv or send is not None")
      self.connect = connect
      self._conn = None

  @property
  def local_store(self):
    ''' The current local Store.
    '''
    return self._local_store

  @local_store.setter
  def local_store(self, newS):
    ''' Switch out the local Store for a new one.
    '''
    oldS = self._local_store
    if newS is not oldS:
      if newS:
        newS.open()
      self._local_store = newS
      if oldS:
        oldS.close()

  def switch_to(self, export_name):
    ''' Switch the local backend Store to one of the exports.
    '''
    self.local_store = self.exports[export_name]

  def startup(self):
    ''' Start up the StreamStore.
        Open the local_store if not None.
    '''
    super().startup()
    local_store = self.local_store
    if local_store is not None:
      local_store.open()

  def shutdown(self):
    ''' Shut down the StreamStore.
    '''
    with Pfx("SHUTDOWN %s", self):
      conn = self._conn
      if conn:
        conn.shutdown()
        self._conn = None
      local_store = self.local_store
      if local_store is not None:
        local_store.close()
      super().shutdown()

  def connection(self):
    ''' Return the current connection, creating it if necessary.
        Returns None if there was no current connection
        and it is too soon since the last connection attempt.
    '''
    with self._lock:
      conn = self._conn
      if conn is None:
        next_attempt = self._conn_attempt_last + self._conn_attempt_delay
        now = time.time()
        if now >= next_attempt:
          self._conn_attemp_last = now
          try:
            recv, send = self.connect()
          except Exception as e:
            error("connect fails: %s: %s", type(e).__name__, e)
          else:
            self._conn = conn = self._packet_connection(recv, send)
            # arrange to disassociate if the channel goes away
            conn.notify_recv_eof.add(self._packet_disconnect)
            conn.notify_send_eof.add(self._packet_disconnect)
    return conn

  def _packet_connection(self, recv, send):
    ''' Wrap a pair of binary streams in a PacketConnection.
    '''
    conn = PacketConnection(
        recv, send, self._handle_request,
        name='PacketConnection:'+self.name)
    return conn

  @locked
  def _packet_disconnect(self, conn):
    warning("PacketConnection DISCONNECT notification: %s", conn)
    oconn = self._conn
    if oconn is conn:
      self._conn = None
      self._conn_attemp_last = time.time()
    else:
      warning("disconnect of %s, but that is not the current connection, ignoring", conn)

  def join(self):
    ''' Wait for the PacketConnection to shut down.
    '''
    conn = self._conn
    if conn:
      conn.join()

  def do(self, rq):
    ''' Wrapper for self._conn.do to catch and report failed autoconnection.
    '''
    with Pfx("%s.do(%s)", self, rq):
      conn = self.connection()
      if conn is None:
        raise StoreError("no connection")
      try:
        retval = conn.do(rq.RQTYPE, rq.flags, bytes(rq))
      except ClosedError as e:
        self._conn = None
        raise StoreError("connection closed: %s" % (e,)) from e
      except CancellationError as e:
        raise StoreError("request cancelled: %s" % (e,)) from e
      else:
        if retval is None:
          raise StoreError("NO RESPONSE from %s" % (rq,))
      return retval

  @staticmethod
  def decode_request(rq_type, flags, payload):
    ''' Decode `(flags, payload)` into a request packet.
    '''
    with Pfx(
        "decode_request(rq_type=%s, flags=0x%02x, payload=%d bytes)",
        rq_type, flags, len(payload)
    ):
      request_class = RqType(rq_type).request_class
      payload_bfr = CornuCopyBuffer.from_bytes(payload)
      rq = request_class.from_buffer(payload_bfr, flags=flags)
      if payload_bfr.offset < len(payload):
        warning(
            "%d unparsed bytes remaining in payload",
            len(payload) - payload_bfr.offset)
      return rq

  @logexc
  def _handle_request(self, rq_type, flags, payload):
    ''' Perform the action for a request packet.
        Return as for the `request_handler` parameter to `PacketConnection`.
    '''
    local_store = self._local_store
    if local_store is None:
      raise ValueError("no local_store, request rejected")
    rq = self.decode_request(rq_type, flags, payload)
    return rq.do(self)

  def add(self, data):
    hashclass = self.hashclass
    h = hashclass.from_chunk(data)
    if self.mode_addif:
      if self.contains(h):
        return h
    ok, flags, payload = self.do(AddRequest(data))
    if not ok:
      raise StoreError(
          "NOT OK response from add(data=%d bytes): flags=0x%0x, payload=%r"
          % (len(data), flags, payload))
    h2, offset = hash_decode(payload)
    if offset != len(payload):
      raise StoreError("extra payload data after hashcode: %r" % (payload[offset:],))
    assert flags == 0
    if h != h2:
      raise RuntimeError(
          "hashclass=%s: precomputed hash %s:%s != hash from .add %s:%s"
          % (hashclass, type(h), h, type(h2), h2))
    return h

  def get(self, h):
    ok, flags, payload = self.do(GetRequest(h))
    if not ok:
      raise StoreError(
          "NOT OK response from get(h=%s): flags=0x%0x, payload=%r"
          % (h, flags, payload))
    found = flags & 0x01
    if found:
      flags &= ~0x01
    if flags:
      raise StoreError("unexpected flags: 0x%02x" % (flags,))
    if found:
      return payload
    if payload:
      raise ValueError("not found, but payload=%r" % (payload,))
    return None

  def contains(self, h):
    ''' Dispatch a contains request, return a Result for collection.
    '''
    ok, flags, payload = self.do(ContainsRequest(h))
    if not ok:
      raise ValueError(
          "NOT OK response from contains(h=%s): flags=0x%0x, payload=%r"
          % (h, flags, payload))
    found = flags & 0x01
    if found:
      flags &= ~0x01
    if flags:
      raise StoreError("unexpected flags: 0x%02x" % (flags,))
    if payload:
      raise StoreError("unexpected payload: %r" % (payload,))
    return found

  def flush(self):
    _, flags, payload = self.do(FlushRequest())
    assert flags == 0
    assert not payload
    local_store = self.local_store
    if local_store is not None:
      local_store.flush()

  def hashcodes_missing(self, other, window_size=None):
    ''' Generator yielding hashcodes in `other` which are missing in `self`.
    '''
    return missing_hashcodes_by_checksum(self, other, window_size=window_size)

  def hashcodes(
      self,
      start_hashcode=None,
      reverse=False, after=False, length=None
  ):
    if length is not None and length < 1:
      raise ValueError("length should be None or >1, got: %r" % (length,))
    if after and start_hashcode is None:
      raise ValueError("after=%s but start_hashcode=%s" % (after, start_hashcode))
    if length is not None and length < 1:
      raise ValueError("length should be None or >1, got: %r" % (length,))
    if after and start_hashcode is None:
      raise ValueError("after=%s but start_hashcode=%s" % (after, start_hashcode))
    ok, flags, payload = self.do(HashCodesRequest(
        hashclass=self.hashclass,
        start_hashcode=start_hashcode,
        reverse=reverse, after=after, length=length))
    if not ok:
      raise StoreError(
          "NOT OK response from hashcodes(h=%s): flags=0x%0x, payload=%r"
          % (start_hashcode, flags, payload))
    if flags:
      raise StoreError("unexpected flags: 0x%02x" % (flags,))
    offset = 0
    hashary = []
    while offset < len(payload):
      hashcode, offset = hash_decode(payload, offset)
      hashary.append(hashcode)
    return hashary

  def hash_of_hashcodes(
      self,
      hashclass=None,
      start_hashcode=None,
      reverse=False, after=False, length=None
  ):
    if length is not None and length < 1:
      raise ValueError("length should be None or >1, got: %r" % (length,))
    if after and start_hashcode is None:
      raise ValueError("after=%s but start_hashcode=%s" % (after, start_hashcode))
    if hashclass is None:
      hashclass = self.hashclass
    ok, flags, payload = self.do(HashOfHashCodesRequest(
        hashclass=hashclass,
        start_hashcode=start_hashcode,
        reverse=reverse, after=after, length=length))
    if not ok:
      raise StoreError(
          "NOT OK response from hash_of_hashcodes: flags=0x%0x, payload=%r"
          % (flags, payload))
    if flags:
      raise StoreError("unexpected flags: 0x%02x" % (flags,))
    hashcode, offset = hash_decode(payload, 0)
    if offset == len(payload):
      h_final = None
    else:
      h_final, offset = hash_decode(payload, offset)
      if offset < len(payload):
        raise StoreError(
            "after hashcode (%s) and h_final (%s), extra bytes: %r"
            % (hashcode, h_final, payload[offset:]))
    return hashcode, h_final

  def hashcodes_from(self, start_hashcode=None, reverse=False):
    ''' Unbounded sequence of hashcodes obtained by successive calls to self.hashcodes.
    '''
    length = 64
    after = False
    while True:
      hashcodes = self.hashcodes(start_hashcode=start_hashcode,
                                 reverse=reverse,
                                 after=after,
                                 length=length)
      if not hashcodes:
        return
      for hashcode in hashcodes:
        yield hashcode
      # set the resume point: after the last yielded hashcode
      start_hashcode = hashcode
      after = True

####################################################################
# Request packet definitions here.
# The RqType enum is defined below these, associating the packet
# classes with each RqType value.
#

class VTPacket(PacketField):
  ''' Base packet class for VT stream requests and responses.

      The real stream packet has the VTPacket.flags value folded
      into its flags.
  '''

  def __init__(self, value):
    super().__init__(value)
    self.flags = 0

class AddRequest(VTPacket):
  ''' An add(bytes) request, returning the hashcode for the stored data.
  '''

  RQTYPE = RqType.ADD

  def __str__(self):
    return "%s(%d:%r...)" % (type(self).__name__, len(self.value), self.value[:16])

  @classmethod
  def value_from_buffer(cls, bfr, flags=0):
    if flags:
      raise ValueError("flags should be 0x00, received 0x%02x" % (flags,))
    return b''.join(bfr)

  def transcribe(self):
    ''' Return the payload as is.
    '''
    return self.value

  def do(self, stream):
    ''' Add data to the local store, return hashcode.
    '''
    local_store = stream._local_store
    if local_store is None:
      raise ValueError("no local_store, request rejected")
    return local_store.add(self.value).encode()

class GetRequest(VTPacket):
  ''' A get(hashcode) request, returning the associated bytes.
  '''

  RQTYPE = RqType.GET

  @staticmethod
  def value_from_buffer(bfr, flags=0):
    if flags:
      raise ValueError("flags should be 0x00, received 0x%02x" % (flags,))
    return hash_from_buffer(bfr)

  def transcribe(self):
    ''' Return the serialised hashcode.
    '''
    return self.value.encode()

  def do(self, stream):
    ''' Return data from the local store by hashcode.
    '''
    local_store = stream._local_store
    if local_store is None:
      raise ValueError("no local_store, request rejected")
    hashcode = self.value
    data = local_store.get(hashcode)
    if data is None:
      return 0
    return 1, data

class ContainsRequest(VTPacket):
  ''' A request to test for the presence of a hashcode.
  '''

  RQTYPE = RqType.CONTAINS

  @staticmethod
  def value_from_buffer(bfr, flags=0):
    if flags:
      raise ValueError("flags should be 0x00, received 0x%02x" % (flags,))
    return hash_from_buffer(bfr)

  def transcribe(self):
    ''' Return the serialised hashcode.
    '''
    return self.value.encode()

  def do(self, stream):
    ''' Return data from the local store by hashcode.
    '''
    local_store = stream._local_store
    if local_store is None:
      raise ValueError("no local_store, request rejected")
    hashcode = self.value
    return 1 if hashcode in local_store else 0

class FlushRequest(VTPacket):
  ''' A flush request.
  '''

  RQTYPE = RqType.FLUSH

  def __init__(self):
    super().__init__(None)

  @staticmethod
  def value_from_buffer(bfr, flags=0):
    if flags:
      raise ValueError("flags should be 0x00, received 0x%02x" % (flags,))
    return None

  @staticmethod
  def do(stream):
    ''' Return data from the local store by hashcode.
    '''
    local_store = stream._local_store
    if local_store is None:
      raise ValueError("no local_store, request rejected")
    local_store.flush()

class HashCodesRequest(Packet):
  ''' A request for remote hashcodes.
  '''

  RQTYPE = RqType.HASHCODES

  def __init__(
      self,
      reverse=False, after=False,
      hashclass=None,
      start_hashcode=None,
      length=None,
  ):
    assert isinstance(reverse, bool)
    assert isinstance(after, bool)
    if hashclass is None:
      raise ValueError("missing hashclass")
    if length is None:
      length = 0
    if after and start_hashcode is None:
      raise ValueError(
          "after=%s but start_hashcode=%s"
          % (after, start_hashcode))
    self.reverse = reverse
    self.after = after
    super().__init__(
        hashname=BSString(hashclass.HASHNAME),
        start_hashcode=(
            HashCodeField(start_hashcode)
            if start_hashcode is not None
            else EmptyField),
        length=BSUInt(length),
    )

  @property
  def flags(self):
    ''' Compute the flags for the request.
    '''
    return (
        ( 0x01 if self.reverse else 0x00 )
        | ( 0x02 if self.after else 0x00 )
        | ( 0x04 if self.start_hashcode is not None else 0x00 )
    )

  @classmethod
  def from_buffer(cls, bfr, flags=0):
    ''' Parse a HashCodesRequest from a buffer and construct.
    '''
    reverse = (flags & 0x01) != 0
    after = (flags & 0x02) != 0
    has_start_hashcode = (flags & 0x04) != 0
    extra_flags = flags & ~0x07
    if extra_flags:
      raise ValueError("extra flags: 0x%02x" % (extra_flags,))
    hashname = BSString.value_from_buffer(bfr)
    hashclass = HASHCLASS_BY_NAME[hashname]
    if has_start_hashcode:
      start_hashcode = HashCodeField.value_from_buffer(bfr)
      if type(start_hashcode) is not hashclass:
        raise ValueError(
            "request hashclass %s does not match start_hashcode class %s"
            % (hashclass, type(start_hashcode)))
    else:
      start_hashcode = None
    length = BSUInt.value_from_buffer(bfr)
    return cls(
        reverse=reverse, after=after,
        hashclass=hashclass,
        start_hashcode=start_hashcode,
        length=length)

  def do(self, stream):
    ''' Return hashcodes from the local store.
    '''
    local_store = stream._local_store
    if local_store is None:
      raise ValueError("no local_store, request rejected")
    # return joined encoded hashcodes
    length = self.length
    if length == 0:
      length = None
    return b''.join(
        h.encode() for h in local_store.hashcodes(
            start_hashcode=self.start_hashcode,
            reverse=self.reverse,
            after=self.after,
            length=length))

class HashOfHashCodesRequest(HashCodesRequest):
  ''' A request for a hashcode of remote hashcodes.
  '''

  RQTYPE = RqType.HASHCODES_HASH

  def do(self, stream):
    ''' Return a hash of hashcodes from the local store.
    '''
    local_store = stream._local_store
    if local_store is None:
      raise ValueError("no local_store, request rejected")
    over_hashcode, final_hashcode = local_store.hash_of_hashcodes(
        start_hashcode=self.start_hashcode,
        reverse=self.reverse,
        after=self.after,
        length=self.length)
    payload = over_hashcode.encode()
    if final_hashcode:
      payload += final_hashcode.encode()
    return payload

RqType.ADD.request_class = AddRequest
RqType.GET.request_class = GetRequest
RqType.CONTAINS.request_class = ContainsRequest
RqType.FLUSH.request_class = FlushRequest
RqType.HASHCODES.request_class = HashCodesRequest
RqType.HASHCODES_HASH.request_class = HashOfHashCodesRequest

if __name__ == '__main__':
  from .stream_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python -tt
#
# Import the C buffer scanner, building it if necessary.
# Fall back to the pure python one if we fail.
#   - Cameron Simpson <cs@cskk.id.au>
#

''' The byte scanning function scanbuf.
    In C by choice, in Python if not.
'''

from distutils.core import setup, Extension
from os import chdir, getcwd
from os.path import dirname, join as joinpath
import sys
##from time import sleep
from cs.logutils import error, warning
from cs.x import X

try:
  from ._scan import scanbuf
except ImportError:
  warning("building _scan from _scan.c")
  def do_setup():
    ''' Run distutils.core.setup from the top of the lib tree.
        Side effect: changes directory, needs undoing.
    '''
    pkgdir = dirname(__file__)
    chdir(dirname(dirname(pkgdir)))
    return setup(
        ext_modules=[Extension("cs.vt._scan", [joinpath(pkgdir, '_scan.c')])],
    )
  ### delay, seemingly needed to make the C version look "new"
  ##sleep(2)
  oargv = sys.argv
  owd = getcwd()
  sys.argv = [oargv[0], 'build_ext', '--inplace']
  try:
    do_setup()
  except SystemExit as e:
    chdir(owd)
    error("SETUP FAILS: %s:%s", type(e), e)
    scanbuf = None
  else:
    chdir(owd)
    try:
      from ._scan import scanbuf
    except ImportError as e:
      error("import fails after setup: %s", e)
      scanbuf = None
  finally:
    sys.argv = oargv

if scanbuf is None:
  warning("using pure Python scanbuf")
  def scanbuf(hash_value, chunk):
    ''' Pure Python scanbuf if there's no C version.
    '''
    offsets = []
    for offset, b in enumerate(chunk):
      hash_value = (
          ( ( hash_value & 0x001fffff ) << 7 )
          | ( ( b & 0x7f )^( (b & 0x80)>>7 ) )
      )
      if hash_value % 4093 == 4091:
        offsets.append(offset)
    return hash_value, offsets

if False:
  # debugging wrapper
  scanbuf0 = scanbuf
  def scanbuf(h, data):
    ''' Debugging scan function.
    '''
    X("scan %d bytes", len(data))
    h2, offsets = scanbuf0(h, data)
    ##X("scan => %r", offsets)
    return h2, offsets



#!/usr/bin/python
#
# A cache store, connected to a fast cache and a slower backend.
#       - Cameron Simpson <cs@cskk.id.au> 07dec2007
#

''' Caching Stores and associated data structures.
'''

from __future__ import with_statement
from collections import namedtuple
import sys
from tempfile import TemporaryFile
from threading import Thread
from cs.fileutils import RWFileBlockCache, datafrom_fd
from cs.logutils import error, warning
from cs.queues import IterableQueue
from cs.resources import RunState, RunStateMixin
from cs.result import Result
from cs.x import X
from . import defaults, MAX_FILE_SIZE, Lock, RLock
from .hash import DEFAULT_HASHCLASS
from .store import BasicStoreSync, MappingStore

DEFAULT_CACHEFILE_HIGHWATER = MAX_FILE_SIZE
DEFAULT_MAX_CACHEFILES = 3

class FileCacheStore(BasicStoreSync):
  ''' A Store wrapping another Store that provides fast access to
      previously fetched data and fast storage of new data.
      asynchronous updates to the backing Store (which may be None).

      This class is a thin Store shaped shim over a FileDataMappingProxy,
      which does the heavy lifting of storing data.
  '''

  def __init__(
      self,
      name, backend, dirpath,
      max_cachefile_size=None, max_cachefiles=None,
      runstate=None,
      hashclass=None,
      **kw
  ):
    ''' Initialise the FileCacheStore.

        Parameters:
        * `name`: the Store name
        * `backend`: the backing Store; this may be None, and the
          property .backend may be switched to another Store at any
          time
        * `dirpath`: directory to hold the cache files
        * `hashclass`: hash class for data chunks;
          if the `backend`
    '''
    if backend:
      backend.open()
      if hashclass is None:
        hashclass = backend.hashclass
      elif hashclass is not backend.hashclass:
        raise ValueError(
            "supplied hashclass %s does not match backend hashclass %s from %s"
            % (hashclass, backend.hashclass, backend))
    else:
      hashclass = DEFAULT_HASHCLASS
      warning("%s:%r: using default hashclass: %s", type(self), name, hashclass)
    super().__init__(name, runstate=runstate, **kw)
    self._attrs.update(backend=backend)
    self._backend = backend
    self.hashclass = hashclass
    self.cache = FileDataMappingProxy(
        backend, dirpath=dirpath,
        max_cachefile_size=max_cachefile_size,
        max_cachefiles=max_cachefiles,
        runstate=runstate,
    )
    self._attrs.update(
        cachefiles=self.cache.max_cachefiles,
        cachesize=self.cache.max_cachefile_size
    )

  def __getattr__(self, attr):
    return getattr(self.backend, attr)

  @property
  def backend(self):
    ''' Return the current backend Store.
    '''
    return self._backend

  @backend.setter
  def backend(self, new_backend):
    ''' Switch backends.
    '''
    old_backend = self._backend
    if old_backend is not new_backend:
      if old_backend:
        old_backend.close()
      self._backend = new_backend
      self.cache.backend = new_backend
      self._attrs.update(backend=new_backend)
      if new_backend:
        new_backend.open()

  def shutdown(self):
    self.cache.close()
    if self.backend:
      self.backend.close()
    super().shutdown()

  def flush(self):
    ''' Dummy flush operation.
    '''
    pass

  def sync(self):
    ''' Dummy sync operation.
    '''
    pass

  def keys(self):
    return self.cache.keys()

  def contains(self, h):
    return h in self.cache

  def add(self, data):
    h = self.hashclass.from_chunk(data)
    self.cache[h] = data
    return h

  # add is deliberately very fast; just return a completed Result directly
  def add_bg(self, data):
    return Result(result=self.add(data))

  def get(self, h):
    try:
      data = self.cache[h]
    except KeyError:
      data = None
    else:
      pass
    return data

_CachedData = namedtuple('CachedData', 'cachefile offset length')
class CachedData(_CachedData):
  ''' A CachedData record, with cachefile, offset, length and implied data.
  '''
  def fetch(self):
    ''' Fetch the data associated with this CachedData instance.
    '''
    return self.cachefile.get(self.offset, self.length)

class FileDataMappingProxy(RunStateMixin):
  ''' Mapping-like class to cache data chunks to bypass gdbm indices and the like.
      Data are saved immediately into an in memory cache and an asynchronous
      worker copies new data into a cache file and also to the backend
      storage.
  '''

  def __init__(
      self, backend, dirpath=None,
      max_cachefile_size=None, max_cachefiles=None,
      runstate=None,
  ):
    ''' Initialise the cache.
        `backend`: mapping underlying us
        `dirpath`: directory to store cache files
        `max_cachefile_size`: maximum cache file size; a new cache
          file is created if this is exceeded; default:
          DEFAULT_CACHEFILE_HIGHWATER
        `max_cachefiles`: number of cache files to keep around; no
          more than this many cache files are kept at a time; default:
          DEFAULT_MAX_CACHEFILES
    '''
    RunStateMixin.__init__(self, runstate=runstate)
    if max_cachefile_size is None:
      max_cachefile_size = DEFAULT_CACHEFILE_HIGHWATER
    if max_cachefiles is None:
      max_cachefiles = DEFAULT_MAX_CACHEFILES
    self.backend = backend
    self.dirpath = dirpath
    self.max_cachefile_size = max_cachefile_size
    self.max_cachefiles = max_cachefiles
    self.cached = {}    # map h => data
    self.saved = {}     # map h => _CachedData(cachefile, offset, length)
    self._lock = Lock()
    self.cachefiles = []
    self._add_cachefile()
    self._workQ = IterableQueue()
    self._worker = Thread(target=self._work)
    self._worker.start()
    self.runstate.notify_cancel.add(lambda rs: self.close())

  def close(self):
    ''' Shut down the cache.
        Stop the worker, close the file cache.
    '''
    self._workQ.close()
    self._worker.join()
    if self.cached:
      error("blocks still in memory cache: %r", self.cached)
    for cachefile in self.cachefiles:
      cachefile.close()

  def _add_cachefile(self):
    cachefile = RWFileBlockCache(dirpath=self.dirpath)
    self.cachefiles.insert(0, cachefile)
    if len(self.cachefiles) > self.max_cachefiles:
      old_cachefile = self.cachefiles.pop()
      old_cachefile.close()

  def _getref(self, h):
    ''' Fetch a cache reference from self.saved, return None if missing.
        Automatically prune stale saved entries if the cachefile is closed.
    '''
    saved = self.saved
    ref = saved.get(h)
    if ref is not None:
      if ref.cachefile.closed:
        ref = None
        del saved[h]
    return ref

  def __contains__(self, h):
    ''' Mapping method supporting "in".
    '''
    with self._lock:
      if h in self.cached:
        return True
      if self._getref(h) is not None:
        return True
    backend = self.backend
    if backend:
      return h in backend
    return False

  def keys(self):
    ''' Mapping method for .keys.
    '''
    seen = set()
    for h in self.cached:
      yield h
      seen.add(h)
    saved = self.saved
    for h in saved:
      if h not in seen and self._getref(h):
        yield h
        seen.add(h)
    backend = self.backend
    if backend:
      for h in backend.keys():
        if h not in seen:
          yield h

  def __getitem__(self, h):
    ''' Fetch the data with key `h`. Raise KeyError if missing.
    '''
    with self._lock:
      # fetch from memory
      try:
        data = self.cached[h]
      except KeyError:
        # fetch from file
        ref = self._getref(h)
        if ref is not None:
          return ref.fetch()
      else:
        # straight from memory cache
        return data
    # not in memory or file cache: fetch from backend, queue store into cache
    backend = self.backend
    if not backend:
      raise KeyError('no backend: h=%s' % (h,))
    data = backend[h]
    with self._lock:
      self.cached[h] = data
    self._workQ.put( (h, data, False) )
    return data

  def __setitem__(self, h, data):
    ''' Store `data` against key `h`.
    '''
    with self._lock:
      if h in self.cached:
        # in memory cache, do not save
        return
      if self._getref(h):
        # in file cache, do not save
        return
      # save in memory cache
      self.cached[h] = data
    # queue for file cache and backend
    self._workQ.put( (h, data, True) )

  def _work(self):
    for h, data, in_backend in self._workQ:
      with self._lock:
        if self._getref(h):
          # already in file cache, therefore already sent to backend
          return
      cachefile = self.cachefiles[0]
      offset = cachefile.put(data)
      with self._lock:
        self.saved[h] = CachedData(cachefile, offset, len(data))
        # release memory cache entry
        try:
          del self.cached[h]
        except KeyError:
          pass
        # roll over to new cache file
        if offset + len(data) >= self.max_cachefile_size:
          self._add_cachefile()
      # store into the backend
      if not in_backend:
        backend = self.backend
        if backend:
          self.backend[h] = data

def MemoryCacheStore(name, max_data):
  ''' Factory to make a MappingStore of a MemoryCacheMapping.
  '''
  return MappingStore(name, MemoryCacheMapping(max_data))

class MemoryCacheMapping:
  ''' A lossy MT-safe in-memory mapping of hashcode->data.
  '''

  def __init__(self, max_data, lock=None):
    if max_data < 65536:
      raise ValueError("max_data should be >= 65536, got: %s" % (max_data,))
    if lock is None:
      lock = Lock()
    self._lock = lock
    self.max_data = max_data
    self.used_data = 0
    self.mapping = {}
    self._ticker = 0
    self._tick = {}
    self._skip_flush = 32

  def __str__(self):
    return (
        "%s[max_data=%d:used_data=%d:hashcodes=%d]"
        % (type(self).__name__, self.max_data, self.used_data, len(self.mapping))
    )

  def __len__(self):
    return len(self.mapping)

  def __contains__(self, hashcode):
    mapping = self.mapping
    with self._lock:
      return hashcode in mapping

  def __getitem__(self, hashcode):
    mapping = self.mapping
    with self._lock:
      data = mapping[hashcode]
      self._tick[hashcode] = self._ticker
      self._ticker += 1
    return data

  def get(self, hashcode, default=None):
    ''' Get the data associated with `hashcode`, or `default`.
    '''
    mapping = self.mapping
    with self._lock:
      try:
        data = mapping[hashcode]
      except KeyError:
        return default
      self._tick[hashcode] = self._ticker
      self._ticker += 1
    return data

  def __setitem__(self, hashcode, data):
    mapping = self.mapping
    with self._lock:
      if hashcode in mapping:
        # DEBUG: sanity check
        if data != mapping[hashcode]:
          raise RuntimeError(
              "data mismatch: hashcode=%s, data=%r vs mapping[hashcode]=%r"
              % (hashcode, data, mapping[hashcode]))
      else:
        mapping[hashcode] = data
        used_data = self.used_data = self.used_data + len(data)
        self._tick[hashcode] = self._ticker
        self._ticker += 1
        max_data = self.max_data
        if used_data > max_data and len(mapping) > 1:
          if self._skip_flush > 0:
            self._skip_flush -= 1
          else:
            _tick = self._tick
            for _, old_hashcode in sorted( (tick, h) for h, tick in _tick.items() ):
              old_data = mapping.pop(old_hashcode)
              used_data -= len(old_data)
              del _tick[old_hashcode]
              if used_data <= max_data:
                break
            self._skip_flush = 32

class BlockMapping:
  ''' A Block's contents mapped onto a temporary file.
  '''

  def __init__(self, tempf, offset, size):
    ''' Initialise the mapping.

        Parameters:
        * `tempf`: the BlockTempfile
        * `offset`: where the Block data start
        * `size`: the total size of the Block
    '''
    self.tempf = tempf
    self.offset = offset
    self.size = size
    self.filled = 0

  def pread(self, size, offset):
    ''' Return data from the main tempfile.
    '''
    assert offset >= 0
    assert offset + size <= self.filled
    return self.tempf.pread(size, self.offset + offset)

  def datafrom(self, offset=0, maxlength=None):
    ''' Yield data from the underlying temp file.

        Parameters:
        * `offset`: start of data, default `0`
        * `maxlength`: maximum amount of data to yield,
          default `self.filled - offset`
    '''
    assert offset >= 0
    if maxlength is None:
      maxlength = self.filled - offset
    else:
      maxlength = min(maxlength, self.filled - offset)
    if maxlength <= 0:
      return
    yield from datafrom_fd(
        self.tempf.fileno(),
        self.offset + offset,
        maxlength=maxlength)

class BlockTempfile:
  ''' Manage a temporary file which contains the contents of various Blocks.
  '''

  def __init__(self, cache, tmpdir, suffix):
    X("new BlockTemptfile...")
    self.cache = cache
    self.tempfile = TemporaryFile(dir=tmpdir, suffix=suffix)
    self.hashcodes = {}
    self.size = 0
    self._lock = RLock()

  def close(self):
    ''' Release the tempfile and unmap the associates hashcodes.
    '''
    blockmap = self.cache.blockmap
    for h in self.hashcodes:
      del blockmap[h]
    self.hashcodes = None
    self.tempfile.close()
    self.tempfile = None

  def pread(self, size, offset):
    ''' Read `size` bytes from the temp file at `offset`.
    '''
    tempf = self.tempfile
    with self._lock:
      if tempf.tell() != offset:
        tempf.seek(offset)
      return tempf.read(size)

  def _pwrite(self, data, offset):
    ''' Write data into the tempfile.

        This is not public as these files are read only.
    '''
    tempf = self.tempfile
    with self._lock:
      if tempf.tell() != offset:
        tempf.seek(offset)
      return tempf.write(data)

  def append_block(self, block, runstate):
    ''' Add a Block to this tempfile.

        Parameters:
        * `block`: the Block to append to this tempfile.
        * `runstate`: a RunState that can be used to cancel the
          tempfile data population Thread

        A Thread is dispatched to load the Block data into the temp file.
    '''
    X("BlockTempfile.append_block(%s)...", block.hashcode)
    h = block.hashcode
    bsize = len(block)
    assert bsize > 0
    with self._lock:
      offset = self.size
      self._pwrite(b'\0', offset + bsize - 1)
      self.size = offset + bsize
    bm = BlockMapping(self.tempfile, offset, bsize)
    with self._lock:
      self.hashcodes[h] = bm
    T = Thread(
        name="%s._infill(%s)" % (type(self).__name__, block),
        target=self._infill, args=(defaults.S, bm, offset, block, runstate))
    T.daemon = True
    T.start()
    return bm

  def _infill(self, S, bm, offset, block, runstate):
    ''' Load the Block data into the tempfile,
        updating the `BlockMapping.filled` attribute as we go.
    '''
    with S:
      needed = len(block)
      for data in block.datafrom():
        if runstate.cancelled:
          break
        assert len(data) <= needed
        written = self._pwrite(data, offset)
        assert written == len(data)
        offset += written
        needed -= written
        bm.filled += written
    X("BlockTempfile._infill(%s) COMPLETE", block.hashcode)

class BlockCache:
  ''' A temporary file based cache for whole Blocks.

      This is to support filesystems' and files' direct read/write
      actions by passing them straight through to this cache is
      there's a mapping.

      We accrue complete Block contents in unlinked files.
  '''

  # default cace size
  MAX_FILES = 32
  MAX_FILE_SIZE = 1024 * 1024 * 1024

  def __init__(self, tmpdir=None, suffix='.dat', max_files=None, max_file_size=None):
    if max_files is None:
      max_files = self.MAX_FILES
    if max_file_size is None:
      max_file_size = self.MAX_FILE_SIZE
    self.tmpdir = tmpdir
    self.suffix = suffix
    self.max_files = max_files
    self.max_file_size = max_file_size
    self.blockmaps = {}     # hashcode -> BlockMapping
    self._tempfiles = []    # in play BlockTempfiles
    self._lock = Lock()
    self.runstate = RunState()
    self.runstate.start()

  def close(self):
    ''' Release all the blockmaps.
    '''
    self.runstate.cancel()
    with self._lock:
      self.blockmaps = {}
      for tempf in self._tempfiles:
        tempf.close()
      self._tempfiles = []

  def __getitem__(self, hashcode):
    ''' Fetch BlockMapping associated with `hashcode`, raise KeyError if missing.
    '''
    return self.blockmaps[hashcode]

  def get_blockmap(self, block):
    ''' Add the specified Block to the cache, return the BlockMapping.
    '''
    blockmaps = self.blockmaps
    h = block.hashcode
    with self._lock:
      bm = blockmaps.get(h)
      if bm is None:
        tempfiles = self._tempfiles
        if tempfiles:
          tempf = tempfiles[-1]
          if tempf.size >= self.max_file_size:
            tempf = None
        else:
          tempf = None
        if tempf is None:
          while len(tempfiles) >= self.max_files:
            otempf = tempfiles.pop(0)
            otempf.close()
          tempf = BlockTempfile(self, self.tmpdir, self.suffix)
          tempfiles.append(tempf)
        bm = tempf.append_block(block, self.runstate)
        blockmaps[h] = bm
    return bm

if __name__ == '__main__':
  from .cache_tests import selftest
  selftest(sys.argv)



#!/usr/bin/env python3
#

''' Functions and classes around hashcodes.
'''

from binascii import unhexlify
from bisect import bisect_left
from hashlib import sha1
import sys
from cs.binary import PacketField, BSUInt
from cs.lex import hexify, get_identifier
from cs.resources import MultiOpenMixin
from cs.serialise import put_bs
from .pushpull import missing_hashcodes
from .transcribe import Transcriber, transcribe_s, register as register_transcriber

class MissingHashcodeError(KeyError):
  ''' Subclass of KeyError raised when accessing a hashcode is not present in the Store.
  '''
  def __init__(self, hashcode):
    KeyError.__init__(self, str(hashcode))
    self.hashcode = hashcode
  def __str__(self):
    return "missing hashcode: %s" % (self.hashcode,)

# enums for hash types, used in encode/decode
HASH_SHA1_T = 0

class HashCodeField(PacketField):
  ''' A PacketField for parsing and transcibing hashcodes.
  '''

  @staticmethod
  def value_from_buffer(bfr):
    ''' Decode a serialised hash from the CornuCopyBuffer `bfr`.
    '''
    hashenum = BSUInt.value_from_buffer(bfr)
    if hashenum == HASH_SHA1_T:
      hashcls = Hash_SHA1
    else:
      raise ValueError("unsupported hashenum %d" % (hashenum,))
    return hashcls.from_hashbytes(bfr.take(hashcls.HASHLEN))

  @staticmethod
  def transcribe_value(hashcode):
    ''' Serialise a hashcode.
    '''
    yield BSUInt.transcribe_value(hashcode.HASHENUM)
    yield hashcode

decode_buffer = HashCodeField.value_from_buffer
decode = HashCodeField.value_from_bytes

def hash_of_byteses(bss):
  ''' Compute a `Hash_SHA1` from the bytes of the supplied `hashcodes`.

      This underlies the mechanism for comparing remote Stores.
  '''
  H = sha1()
  for bs in bss:
    H.update(bs)
  return Hash_SHA1.from_chunk(H.digest())

class HashCode(bytes, Transcriber):
  ''' All hashes are bytes subclasses.
  '''

  __slots__ = ()

  transcribe_prefix = 'H'

  def __str__(self):
    return transcribe_s(self)

  def __repr__(self):
    return ':'.join( (self.HASHNAME, hexify(self)) )

  def __eq__(self, other):
    return self.HASHENUM == other.HASHENUM and bytes.__eq__(self, other)

  def __hash__(self):
    return bytes.__hash__(self)

  @staticmethod
  def from_buffer(bfr):
    ''' Decode a hash from a buffer.
    '''
    return HashCodeField.value_from_buffer(bfr)

  def transcribe_b(self):
    ''' Binary transcription of this hash via `cs.binary.PacketField.transcribe_value`.
    '''
    return HashCodeField.transcribe_value(self)

  def encode(self):
    ''' Return the serialised form of this hash object: hash enum plus hash bytes.
        If we ever have a variable length hash function,
        hash bytes will have to include that information.
    '''
    return bytes(HashCodeField(self))

  @classmethod
  def from_hashbytes(cls, hashbytes):
    ''' Factory function returning a Hash_SHA1 object from the hash bytes.
    '''
    if len(hashbytes) != cls.HASHLEN:
      raise ValueError(
          "expected %d bytes, received %d: %r"
          % (cls.HASHLEN, len(hashbytes), hashbytes))
    return cls(hashbytes)

  @classmethod
  def from_chunk(cls, chunk):
    ''' Factory function returning a HashCode object from a data block.
    '''
    hashbytes = cls.HASHFUNC(chunk).digest()
    return cls.from_hashbytes(hashbytes)

  @property
  def hashfunc(self):
    ''' Convenient hook to this Hash's class' .from_chunk method.
    '''
    return self.__class__.from_chunk

  @property
  def filename(self):
    ''' A file basename for files related to this hashcode: {hashcodehex}.{hashtypename}
    '''
    return hexify(self) + '.' + self.HASHNAME

  def transcribe_inner(self, T, fp):
    fp.write(self.HASHNAME)
    fp.write(':')
    fp.write(hexify(self))

  @staticmethod
  def parse_inner(T, s, offset, stopchar, prefix):
    ''' Parse hashname:hashhextext from `s` at offset `offset`.
        Return HashCode instance and new offset.
    '''
    hashname, offset = get_identifier(s, offset)
    if not hashname:
      raise ValueError("missing hashname at offset %d" % (offset,))
    hashclass = HASHCLASS_BY_NAME[hashname]
    if offset >= len(s) or s[offset] != ':':
      raise ValueError("missing colon at offset %d" % (offset,))
    offset += 1
    hexlen = hashclass.HASHLEN * 2
    hashtext = s[offset:offset + hexlen]
    if len(hashtext) != hexlen:
      raise ValueError("expected %d hex digits, found only %d" % (hexlen, len(hashtext)))
    offset += hexlen
    bs = unhexlify(hashtext)
    H = hashclass.from_hashbytes(bs)
    return H, offset

register_transcriber(HashCode)

class Hash_SHA1(HashCode):
  ''' A hash class for SHA1.
  '''
  __slots__ = ()
  HASHFUNC = sha1
  HASHNAME = 'sha1'
  HASHLEN = 20
  HASHENUM = HASH_SHA1_T
  HASHENUM_BS = put_bs(HASHENUM)
  HASHLEN_ENCODED = len(HASHENUM_BS) + HASHLEN

HASHCLASS_BY_NAME = {}

def register_hashclass(klass):
  ''' Register a hash class for lookup elsewhere.
  '''
  hashname = klass.HASHNAME
  if hashname in HASHCLASS_BY_NAME:
    raise ValueError(
        'cannot register hash class %s: hashname %r already registered to %s'
        % (klass, hashname, HASHCLASS_BY_NAME[hashname]))
  HASHCLASS_BY_NAME[hashname] = klass

register_hashclass(Hash_SHA1)

DEFAULT_HASHCLASS = Hash_SHA1

class HashCodeUtilsMixin(object):
  ''' Utility methods for classes which use hashcodes as keys.
      Subclasses will generally override .hashcodes_from, which
        returns an iterator that yields hashcodes until none remains.
        The default implementation presumes the class is iterable and
        that that iteration yields hashcodes; this works for mappings,
        for example. However, because of the need for sorted output
        the default implementation is expensive. A subclass built on
        some kind of database will often have an efficient key iteration
        that can be used instead.
      The other methods include:
        .hashcodes, the big brother of hashcodes_from with more
          options; the default implentation uses .hashcodes_from and
          is roughly as efficient or inefficient. Classes like
          StreamStore provide their own implementation, but this is
          usually not necessary.
        .hash_of_hashcodes, used for comparing Store contents efficiently
        .hashcodes_missing, likewise
  '''

  def hash_of_hashcodes(
      self,
      start_hashcode=None,
      reverse=None, after=False, length=None
  ):
    ''' Return a hash of the hashcodes requested and the last
        hashcode (or None if no hashcodes matched); used for comparing
        remote Stores.
    '''
    if length is not None and length < 1:
      raise ValueError("length < 1: %r" % (length,))
    if after and start_hashcode is None:
      raise ValueError("after=%s but start_hashcode=%s" % (after, start_hashcode))
    hs = list(
        self.hashcodes(
            start_hashcode=start_hashcode,
            reverse=reverse, after=after, length=length))
    if hs:
      h_final = hs[-1]
    else:
      h_final = None
    return hash_of_byteses(hs), h_final

  def hashcodes_missing(self, other, window_size=None):
    ''' Generator yielding hashcodes in `other` which are missing in `self`.
        Note that a StreamStore overrides this with a call to
        missing_hashcodes_by_checksum to reduce bandwidth.
    '''
    return missing_hashcodes(self, other, window_size=window_size)

  def hashcodes_from(self, start_hashcode=None, reverse=False):
    ''' Default generator yielding hashcodes from this object until none remains.
        This implementation starts by fetching and sorting all the
        keys, so for large mappings this implementation is memory
        expensive and also runtime expensive if only a few hashcodes
        are desired.
        `start_hashcode`: starting hashcode - hashcodes are >=`start_hashcode`;
                          if None start the sequences from the smallest
                          hashcode or from the largest if `reverse` is true
        `reverse`: yield hashcodes in reverse order (counting down instead of up).
    '''
    hashclass = self.hashclass
    if start_hashcode is not None:
      if not isinstance(start_hashcode, hashclass):
        raise TypeError("hashclass %s does not match start_hashcode %r"
                        % (hashclass, start_hashcode))
    ks = sorted(hashcode for hashcode in iter(self) if isinstance(hashcode, hashclass))
    if not ks:
      return
    if start_hashcode is None:
      if reverse:
        ndx = len(ks) - 1
      else:
        ndx = 0
    else:
      ndx = bisect_left(ks, start_hashcode)
      if ndx == len(ks):
        # start_hashcode > max hashcode
        if reverse:
          # step back into array
          ndx -= 1
        else:
          # nothing to return
          return
      else:
        # start_hashcode <= max hashcode
        # ==> ks[ndx] >= start_hashcode
        if reverse and ks[ndx] > start_hashcode:
          if ndx > 0:
            ndx -= 1
          else:
            return
    # yield keys until we're not wanted
    while True:
      try:
        hashcode = ks[ndx]
      except IndexError:
        break
      yield hashcode
      if reverse:
        if ndx == 0:
          break
        ndx -= 1
      else:
        ndx += 1

  def hashcodes(self, start_hashcode=None, reverse=False, after=False, length=None):
    ''' Generator yielding up to `length` hashcodes >=`start_hashcode`.
        This relies on .hashcodes_from as the source of hashcodes.
        `start_hashcode`: starting hashcode - hashcodes are >=`start_hashcode`;
                          if None start the sequences from the smallest
                          hashcode or from the largest if `reverse` is true
        `reverse`: yield hashcodes in reverse order (counting down instead of up).
        `after`: skip the first hashcode if it is equal to `start_hashcode`
        `length`: the maximum number of hashcodes to yield
    '''
    if length is not None and length < 1:
      raise ValueError("length < 1: %r" % (length,))
    if after and start_hashcode is None:
      raise ValueError("after=%s but start_hashcode=%s" % (after, start_hashcode))
    # try to short circuit if there are no hashcodes
    try:
      nhashcodes = len(self)
    except TypeError:
      pass
    else:
      if nhashcodes == 0:
        return
    first = True
    for hashcode in self.hashcodes_from(start_hashcode=start_hashcode,
                                        reverse=reverse):
      if first:
        first = False
        if after and hashcode == start_hashcode:
          # skip start_hashcode if present
          continue
      yield hashcode
      if length is not None:
        length -= 1
        if length < 1:
          break

  def hashcodes_bg(self, start_hashcode=None, reverse=None, after=False, length=None):
    ''' Background a hashcodes call.
    '''
    return self._defer(
        self.hashcodes,
        start_hashcode=start_hashcode,
        reverse=reverse, after=after, length=length)

class HashUtilDict(dict, MultiOpenMixin, HashCodeUtilsMixin):
  ''' Simple dict subclass supporting HashCodeUtilsMixin.
  '''

  def __init__(self):
    dict.__init__(self)
    MultiOpenMixin.__init__(self)
    self.hashclass = DEFAULT_HASHCLASS

  def __str__(self):
    return '<%s:%d-entries>' % (self.__class__.__name__, len(self))

  def add(self, data):
    ''' Add `data` to the dict.
    '''
    hashcode = self.hashclass.from_chunk(data)
    self[hashcode] = data
    return hashcode

  def startup(self):
    ''' Dummy method to support unit tests with open/close.
    '''
    pass

  def shutdown(self):
    ''' Dummy method to support unit tests with open/close.
    '''
    pass

if __name__ == '__main__':
  from .hash_tests import selftest
  selftest(sys.argv)



#!/usr/bin/python
#
# Stream tests.
#       - Cameron Simpson <cs@cskk.id.au>
#

''' Tests for StreamStores.
'''

import os
import sys
import unittest
from cs.x import X
from .hash import DEFAULT_HASHCLASS
from .hash_tests import HashUtilDict, _TestHashCodeUtils
from .store import MappingStore
from .store_tests import TestStore
from .stream import StreamStore

def make_stream_store(hashclass, addif):
  ''' Contruct an in-memory remote Store attached with pipes.
  '''
  upstream_rd, upstream_wr = os.pipe()
  downstream_rd, downstream_wr = os.pipe()
  remote_S = StreamStore(
      "stream_tests.make_stream_store.remote_S",
      upstream_rd,
      downstream_wr,
      local_store=MappingStore(
          "stream_tests.make_stream_store.remote_S.local_store",
          HashUtilDict()).open(),
      addif=addif,
      hashclass=hashclass
  )
  S = StreamStore(
      "stream_tests.make_stream_store.S",
      downstream_rd,
      upstream_wr,
      addif=addif,
      hashclass=hashclass
  )
  return S, remote_S

class TestStreamStore(TestStore, unittest.TestCase):
  ''' Test a stream in non-addif mode.
  '''

  hashclass = DEFAULT_HASHCLASS

  def _init_Store(self):
    self.S, self.remote_S = make_stream_store(self.hashclass, addif=False)

  def setUp(self):
    self._init_Store()
    TestStore.setUp(self)
    self.remote_S.open()

  def tearDown(self):
    self.remote_S.close()
    TestStore.tearDown(self)

class TestStreamStoreAddIf(TestStore, unittest.TestCase):
  ''' test a stream in addif mode.
  '''

  hashclass = DEFAULT_HASHCLASS

  def _init_Store(self):
    self.S, self.remote_S = make_stream_store(self.hashclass, addif=True)

  def setUp(self):
    self._init_Store()
    TestStore.setUp(self)
    self.remote_S.open()

  def tearDown(self):
    self.remote_S.close()
    TestStore.tearDown(self)

##@unittest.skip("too noisy")
class TestHashCodeUtilsStreamStore(_TestHashCodeUtils, unittest.TestCase):
  ''' Test HashUtils on a StreamStore on a HashUtilDict.
  '''
  ADDIF_MODE = False
  hashclass = DEFAULT_HASHCLASS
  def MAP_FACTORY(self):
    S, remote_S = make_stream_store(self.hashclass, self.ADDIF_MODE)
    remote_S.open()
    self.remote_S = remote_S
    return S

  def tearDown(self):
    self.remote_S.close()
    _TestHashCodeUtils.tearDown(self)

class TestHashCodeUtilsStreamStoreAddIf(TestHashCodeUtilsStreamStore):
  ''' Test HashUtils on a StreamStore on a HashUtilDict with addif=True.
  '''
  ADDIF_MODE = True

def selftest(argv):
  ''' Run the unit tests.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# TCP and UNIX domain socket stream tests.
# - Cameron Simpson <cs@cskk.id.au>
#

''' TCP and UNIX domain socket stream tests.
'''

import errno
import sys
import unittest
from .hash import HashUtilDict
from .hash_tests import _TestHashCodeUtils
from .socket import TCPStoreServer, TCPClientStore
from .store import MappingStore
from .store_tests import TestStore

BIND_HOST = '127.0.0.1'
BASE_PORT = 9999

def make_tcp_store():
  global BASE_PORT
  base_port = BASE_PORT
  mapping_S = MappingStore("tcp_tests.make_tcp_store.mapping_S", HashUtilDict())
  while True:
    bind_addr = (BIND_HOST, base_port)
    try:
      remote_S = TCPStoreServer(bind_addr, local_store=mapping_S)
    except OSError as e:
      if e.errno == errno.EADDRINUSE:
        base_port += 1
      else:
        raise
    else:
      break
  remote_S.open()
  S = TCPClientStore(None, bind_addr)
  return S, remote_S

class TestTCPStore(TestStore, unittest.TestCase):
  ''' Tests for TCPStoreServer and TCPClientStore.
  '''

  def _init_Store(self):
    self.S, self.remote_S = make_tcp_store()

  def setUp(self):
    self._init_Store()
    TestStore.setUp(self)
    self.remote_S.open()

  def tearDown(self):
    self.remote_S.close()
    TestStore.tearDown(self)

class TestHashCodeUtilsTCPStore(_TestHashCodeUtils, unittest.TestCase):
  ''' Test HashUtils on a TCPStore on a HashUtilDict.
  '''

  def MAP_FACTORY(self):
    ''' Return context manager for this test's resources.
    '''
    S, remote_S = make_tcp_store()
    remote_S.open()
    self.remote_S = remote_S
    return S

  def tearDown(self):
    self.remote_S.close()
    _TestHashCodeUtils.tearDown(self)
    ##debug_object_shell(self, prompt='%s.tearDown> ' % (self._testMethodName,))

def selftest(argv):
  ''' Run the unit tests with `argv`.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python
#
# Blockify tests.
#       - Cameron Simpson <cs@cskk.id.au>
#

''' Unit tests for cs.vt.blockify.
'''

from collections import defaultdict
from itertools import chain
import os
import os.path
import sys
import time
import unittest
from cs.buffer import chunky, CornuCopyBuffer
from cs.fileutils import read_from
from .randutils import rand0, randblock
from cs.x import X
from .blockify import blockify, blocked_chunks_of, \
                      MAX_BLOCKSIZE, DEFAULT_SCAN_SIZE
from .parsers import scan_text, scan_mp3, scan_mp4
from .store import MappingStore

QUICK = len(os.environ.get('QUICK', '')) > 0

def random_blocks(max_size=65536, count=64):
  ''' Generate `count` blocks of random sizes from 1 to `max_size`.
  '''
  for _ in range(count):
    size = rand0(max_size) + 1
    yield randblock(size)

class TestAll(unittest.TestCase):
  ''' All the unit tests.
  '''

  # load my_code from this test suite
  with open(__file__, 'rb') as myfp:
    mycode = myfp.read()

  X("generate random test data")
  if QUICK:
    random_data = list(random_blocks(max_size=1200, count=12))
  else:
    random_data = list(random_blocks(max_size=12000, count=1280))

  def setUp(self):
    ''' Open the source file.
    '''
    self.fp = open(__file__, "rb")

  def tearDown(self):
    ''' Close the source file.
    '''
    self.fp.close()

  def test01scanners(self):
    ''' Test some domain specific data parsers.
    '''
    rand_total = sum(len(chunk) for chunk in self.random_data)
    for scanner in (scan_text, scan_mp3):
      with self.subTest(scanner.__name__):
        input_chunks = self.random_data
        if scanner is scan_mp3:
          if os.path.exists('TEST.mp3'):
            ##X("mp3 parse: replace input data with chunks from TEST.mp3")
            input_chunks = read_from(open('TEST.mp3', 'rb'))
          else:
            ##X("no TEST.mp3 in ".os.getcwd())
            pass
        Q = scanner(CornuCopyBuffer(iter(input_chunks)))
        last_qoffset = 0
        for qoffset in Q:
          self.assertIsInstance(
              qoffset, int,
              'scanner must yield only ints, received %s:%r'
              % (type(qoffset), qoffset))
          self.assertTrue(
              last_qoffset <= qoffset,
              "qoffset %d <= last_qoffset %d" % (qoffset, last_qoffset))
          last_qoffset = qoffset

  def test02blocked_chunks_of(self):
    ''' Blockify some input sources.
    '''
    for scanner in (
        ##None,
        ##scan_text,
        ##scan_mp3,
        scan_mp4,
    ):
      scanner_desc = 'None' if scanner is None else scanner.__name__
      for input_desc, input_chunks in (
          ##('random data', self.random_data),
          ('100 x ' + __file__, [ self.mycode for _ in range(100) ]),
      ):
        testfile = None
        rfp = None
        if scanner is scan_mp3:
          testfile = 'TEST.mp3'
        elif scanner is scan_mp4:
          testfile = 'TEST.mp4'
        if testfile is not None:
          if os.path.exists(testfile):
            X("%s: replace input data with chunks from %s", scanner, testfile)
            rfp = open(testfile, 'rb')
            input_chunks = read_from(rfp, DEFAULT_SCAN_SIZE)
            input_desc = testfile
          else:
            X("%s: no %s in %s", scanner, testfile, os.getcwd())
        with self.subTest("blocked_chunks_of",
                          scanner=scanner_desc,
                          source=input_desc):
          if True:
            source_chunks = input_chunks
            src_total = None
          else:
            source_chunks = list(input_chunks)
            src_total = 0
            for chunk in source_chunks:
              src_total += len(chunk)
            X("%d source chunks, %d bytes in total", len(source_chunks), src_total)
          chunk_total = 0
          nchunks = 0
          all_chunks = []
          start_time = time.time()
          offset = 0
          chunky_scanner = chunky(scanner) if scanner else None
          histogram = defaultdict(int)
          for chunk in blocked_chunks_of(source_chunks, chunky_scanner, histogram=histogram):
            nchunks += 1
            chunk_total += len(chunk)
            all_chunks.append(chunk)
            offset += len(chunk)
            # the pending.flush operation can return short blocks
            ##self.assertTrue(len(chunk) >= MIN_BLOCKSIZE)
            self.assertTrue(len(chunk) <= MAX_BLOCKSIZE,
                            "len(chunk)=%d > MAX_BLOCKSIZE=%d"
                            % (len(chunk), MAX_BLOCKSIZE))
            if src_total is not None:
              self.assertTrue(chunk_total <= src_total,
                              "chunk_total:%d > src_total:%d"
                              % (chunk_total, src_total))
          end_time = time.time()
          X("%s|%s: received %d chunks in %gs, %d bytes at %g B/s",
            input_desc, scanner_desc,
            nchunks, end_time-start_time, chunk_total,
            float(chunk_total) / (end_time-start_time))
          X("    %d offsets from scanner, %d offsets from hash scan",
            histogram['offsets_from_scanner'],
            histogram['offsets_from_hash_scan'])
          if src_total is not None:
            self.assertEqual(src_total, chunk_total)
            self.assertEqual(b''.join(source_chunks),
                             b''.join(all_chunks))
          if False:
            for hits, size in sorted([
                (hits, size)
                for size, hits in histogram.items()
                if isinstance(size, int)
            ]):
              if hits > 1:
                X("size %d: %d", size, hits)
        if rfp is not None:
          rfp.close()
          rfp = None

  def test03blockifyAndRetrieve(self):
    ''' Blockify some data and ensure that the blocks match the data.
    '''
    with MappingStore("TestAll.test00blockifyAndRetrieve", {}):
      data = self.fp.read()
      blocks = list(blockify([data]))
      ##X("blocks=%r", blocks)
      data2 = b''.join( chain(*[ B.datafrom() for B in blocks ]) )
      self.assertEqual(
          len(data), len(data2),
          "data mismatch: len(data)=%d, len(data2)=%d"
          % (len(data), len(data2)))
      self.assertEqual(
          data, data2,
          "data mismatch: data and data2 same length but contents differ")

def selftest(argv):
  ''' Run the unit tests.
  '''
  unittest.main(__name__, None, argv)

if __name__ == '__main__':
  selftest(sys.argv)



#!/usr/bin/python3
#
# Control operations given a Dirent.
# - Cameron Simpson <cs@cskk.id.au> 31may2018
#

''' Control operations, intended to be available to the command
    line and via FTPlike interfaces and via the setxattr('x-vt-control',
    command_string) action in a FUSE implementation.
'''

from getopt import GetoptError
from cs.logutils import exception
from cs.pfx import Pfx
from . import PATHSEP
from .dir import _Dirent
from .transcribe import parse

class Control:
  ''' Command line interface to perform operations on Dirents.
  '''

  def control(self, E, argv):
    ''' Perform a control action `argv` on the Dirent `E`.
    '''
    with Pfx("%s.control(E=%s,argv=%r)", self, E, argv):
      if not argv:
        raise ValueError('empty argv')
      op = argv.pop()
      with Pfx(op):
        try:
          action = getattr(self, 'cmd_' + op)
        except AttributeError as e:
          exception("unknown operation %r: %s", op, e)
          raise ValueError("unknown operation %r: %s" % ( op, e))
        try:
          return action(E, argv)
        except GetoptError as e:
          exception("%s: %s", type(e), e)
          raise ValueError("%s: %s" % (type(e), e))

  @staticmethod
  def cmd_attach(D, argv):
    ''' Attach an arbitrary Dirent to the current directory `D`.
        Usage: attach name dirent_spec
    '''
    if not D.isdir:
      raise GetoptError("not a Dir: %s" % (D,))
    if not argv:
      raise GetoptError("missing name")
    name = argv.pop()
    if not name or PATHSEP in name:
      raise GetoptError(
          "invalid name, may not be empty or contain the separator %r: %r"
          % (PATHSEP, name)
      )
    if not argv:
      raise GetoptError("missing dirent_spec")
    dirent_spec = argv.pop()
    if argv:
      raise GetoptError("extra arguments after dirent_spec: %r" % (argv,))
    try:
      E = parse(dirent_spec)
    except ValueError as e:
      raise GetoptError("parse failure: %r: %s" % (dirent_spec, e)) from e
    if not isinstance(E, _Dirent):
      raise GetoptError("not a Dirent specification: %r" % (dirent_spec,))
    if name in D:
      raise GetoptError("name already exists: %r" % (name,))
    D[name] = E
